import numpy as np
import scipy.linalg
import scipy.sparse as sparse
import scipy.sparse.linalg as sla
from ._stationary_vector import stationary_distribution


def mfpt(T, target):
    r"""Mean first passage times to a set of target states.

    Parameters
    ----------
    T : sparse matrix or ndarray
        Transition matrix.
    target : int or list of int
        Target states for mfpt calculation.

    Returns
    -------
    m_t : ndarray, shape=(n,)
         Vector of mean first passage times to target states.

    Notes
    -----
    The mean first passage time :math:`\mathbf{E}_x[T_Y]` is the expected
    hitting time of one state :math:`y` in :math:`Y` when starting in state :math:`x`.

    For a fixed target state :math:`y` it is given by

    .. math :: \mathbb{E}_x[T_y] = \left \{  \begin{array}{cc}
                                             0 & x=y \\
                                             1+\sum_{z} T_{x,z} \mathbb{E}_z[T_y] & x \neq y
                                             \end{array}  \right.

    For a set of target states :math:`Y` it is given by

    .. math :: \mathbb{E}_x[T_Y] = \left \{  \begin{array}{cc}
                                             0 & x \in Y \\
                                             1+\sum_{z} T_{x,z} \mathbb{E}_z[T_Y] & x \notin Y
                                             \end{array}  \right.

    References
    ----------
    .. [1] Hoel, P G and S C Port and C J Stone. 1972. Introduction to
        Stochastic Processes.

    Examples
    --------

    >>> from deeptime.markov.tools.analysis import mfpt

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> m_t = mfpt(T, 0)
    >>> m_t
    array([ 0., 12., 22.])

    """
    dim = T.shape[0]
    if sparse.issparse(T):
        solve = sla.spsolve

        A = sparse.eye(dim, dim) - T
        # Convert to DOK (dictionary of keys) matrix to enable row-slicing and assignement
        A = A.todok()
        D = A.diagonal()
        A[target, :] = 0.0
        D[target] = 1.0
        A.setdiag(D)
        # Convert back to CSR-format for fast sparse linear algebra
        A = A.tocsr()
    else:
        solve = scipy.linalg.solve

        A = np.eye(dim) - T
        A[target, :] = 0.0
        A[target, target] = 1.0
    b = np.ones(dim)
    b[target] = 0.0

    m_t = solve(A, b)
    return m_t


def mfpt_between_sets(T, target, origin, mu=None):
    r"""Compute mean-first-passage time between subsets of state space.

    Parameters
    ----------
    T : sparse matrix or ndarray
        Transition matrix.
    target : int or list of int
        Set of target states.
    origin : int or list of int
        Set of starting states.
    mu : (M,) ndarray (optional)
        The stationary distribution of the transition matrix T.

    Returns
    -------
    tXY : float
        Mean first passage time between set X and Y.

    Notes
    -----
    The mean first passage time :math:`\mathbf{E}_X[T_Y]` is the expected
    hitting time of one state :math:`y` in :math:`Y` when starting in a
    state :math:`x` in :math:`X`:

    .. math :: \mathbb{E}_X[T_Y] = \sum_{x \in X}
                \frac{\mu_x \mathbb{E}_x[T_Y]}{\sum_{z \in X} \mu_z}

    """
    if mu is None:
        mu = stationary_distribution(T)

    """Stationary distribution restriced on starting set X"""
    nuX = mu[origin]
    muX = nuX / np.sum(nuX)

    """Mean first-passage time to Y (for all possible starting states)"""
    tY = mfpt(T, target)

    """Mean first-passage time from X to Y"""
    tXY = np.dot(muX, tY[origin])
    return tXY



# .. moduleauthor:: F. Noe <frank DOT noe AT fu-berlin DOT de>
# .. moduleauthor:: B. Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>

from math import ceil
from typing import Optional, List, Union

import numpy as np
import scipy
from scipy.sparse import issparse

from ...base import Model
from ...covariance import CovarianceModel
from ...decomposition import CovarianceKoopmanModel, vamp_score
from .._pcca import pcca, PCCAModel
from .._reactive_flux import ReactiveFlux
from ..sample import ensure_dtraj_list, compute_index_states
from ..tools import analysis as msmana
from .._transition_counting import TransitionCountModel
from .._util import count_states
from ...numeric import is_square_matrix, spd_inv_sqrt
from ...util.decorators import cached_property
from ...util.matrix import submatrix
from ...util.types import ensure_array


class MarkovStateModel(Model):
    r"""Markov model with a given transition matrix.

    Parameters
    ----------
    transition_matrix : (n,n) array_like
        The transition matrix.
    stationary_distribution : ndarray(n), optional, default=None
        Stationary distribution. Can be optionally given in case if it was already computed.
    reversible : bool, optional, default=None
        Whether the transition matrix is reversible with respect to its stationary distribution. If None (default),
        will be determined from the transition matrix.
    n_eigenvalues : int, optional, default=None
        The number of eigenvalues / eigenvectors to be kept. If set to None, it depends on the transition matrix.
        If it is densely stored (in terms of a numpy array), all eigenvectors and eigenvalues are computed. If it is
        sparse, only the 10 largest eigenvalues with corresponding eigenvectors are computed.
    ncv : int optional, default=None
        Relevant for eigenvalue decomposition of reversible transition matrices. It is the number of Lanczos
        vectors generated, `ncv` must be greater than n_eigenvalues; it is recommended that ncv > 2*neig.
    count_model : TransitionCountModel, optional, default=None
        In case the MSM was estimated from data, the transition count model can be provided for statistical
        information about the data. Some properties of the model require a count model so that they can be computed.
    transition_matrix_tolerance : float, default=1e-8
        The tolerance under which a matrix is still considered a transition matrix (only non-negative elements and
        row sums of 1).

    See Also
    --------
    MaximumLikelihoodMSM : maximum-likelihood estimator for MSMs
    OOMReweightedMSM : estimator for MSMs which uses Koopman reweighting
    BayesianMSM : bayesian sampling of MSMs to obtain uncertainties

    References
    ----------
    .. bibliography:: /references.bib
        :style: unsrt
        :filter: docname in docnames
        :keyprefix: msm-
    """

    def __init__(self, transition_matrix, stationary_distribution=None, reversible=None,
                 n_eigenvalues=None, ncv=None, count_model=None, transition_matrix_tolerance=1e-8):
        super().__init__()
        self._is_reversible = reversible
        self._ncv = ncv
        self._transition_matrix_tolerance = transition_matrix_tolerance

        self.update_transition_matrix(transition_matrix)

        if n_eigenvalues is None:
            if self.sparse:
                # expect large matrix, don't take full state space but just (magic) the dominant 10
                n_eigenvalues = min(10, self.n_states - 1)
            else:
                # if we are dense take everything
                n_eigenvalues = self.n_states
        self._n_eigenvalues = n_eigenvalues
        self._count_model = count_model
        self._eigenvalues = None

        self._stationary_distribution = None
        self.update_stationary_distribution(stationary_distribution)

    ################################################################################
    # Basic attributes
    ################################################################################

    def _invalidate_caches(self):
        r""" Invalidates all cached properties and causes them to be re-evaluated """
        for member in MarkovStateModel.__dict__.values():
            if isinstance(member, cached_property):
                member.invalidate()
        self._eigenvalues = None
        self._stationary_distribution = None

    @property
    def count_model(self) -> Optional[TransitionCountModel]:
        r"""
        Returns a transition count model, can be None. The transition count model statistics about data that was used
        for transition counting as well as a count matrix.

        Returns
        -------
        model : TransitionCountModel or None
            The transition count model or None.
        """
        return self._count_model

    @property
    def has_count_model(self) -> bool:
        r""" Yields whether this Markov state model has a count model.

        :type: bool
        """
        return self._count_model is not None

    @cached_property
    def koopman_model(self) -> CovarianceKoopmanModel:
        r"""Yields a :class:`CovarianceKoopmanModel` based on the transition matrix and stationary
         distribution of this Markov state model.

        Returns
        -------
        model : CovarianceKoopmanModel
            The model.

        Notes
        -----
        If :math:`P\in\mathbb{R}^{n\times n}` denotes the transition matrix and :math:`\mu\in\mathbb{R}^n` denotes
        the stationary distribution, we define covariance matrices

        .. math::
            \begin{aligned}
            C_{00} &= C_{11} = \text{diag} (\mu_1,\ldots,\mu_n)\\
            C_{01} &= C_{00}P
            \end{aligned}

        and based on these a Koopman operator :math:`K = C_{00}^{-1/2}C_{01}C_{11}^{-1/2}`.

        See Also
        --------
        to_koopman_model
        """
        return self.to_koopman_model(False)

    @cached_property
    def empirical_koopman_model(self) -> CovarianceKoopmanModel:
        r"""Yields a :class:`CovarianceKoopmanModel` based on the count matrix of this Markov state model.

        Returns
        -------
        model : CovarianceKoopmanModel
            The model.

        Notes
        -----
        If :math:`C\in\mathbb{R}^{n\times n}` denotes the count matrix and :math:`P` the transition matrix,
        we define covariance matrices based on transition count statistics

        .. math::

            \begin{aligned}
            C_{00} &= \text{diag} \left( \sum_i C_{i1}, \ldots, \sum_i C_{in} \right) \\
            C_{11} &= \text{diag} \left( \sum_i C_{1i}, \ldots, \sum_i C_{ni} \right) \\
            C_{01} &= C,
            \end{aligned}

        and reweight the operator :math:`P` to the empirical distribution via :math:`C_{01\text{, re}} = C_{00}P`.
        Based on these we define the Koopman operator :math:`K = C_{00}^{-1/2}C_{01\text{, re}}C_{11}^{-1/2}`.

        See Also
        --------
        to_koopman_model
        """
        return self.to_koopman_model(True)

    def to_koopman_model(self, empirical: bool = True, epsilon: float = 1e-10):
        r""" Computes the SVD of the symmetrized Koopman operator in the analytical or empirical distribution,
        returns as Koopman model.

        Parameters
        ----------
        empirical : bool, optional, default=True
            Determines whether the model should refer to the analytical distributions based on the transition matrix
            or the empirical distributions based on the count matrix.
        epsilon : float, optional, default=1e-10
            Regularization parameter for computing inverses of covariance matrices.

        Returns
        -------
        model : CovarianceKoopmanModel
            The model.
        """
        if empirical:
            assert self.has_count_model, "This requires statistics over the estimation data, " \
                                         "in particular a count model."
        return _svd_sym_koopman(self, empirical, epsilon=epsilon)

    @property
    def lagtime(self) -> int:
        r""" The lagtime this model was estimated at. In case no count model was provided, this property defaults
        to a lagtime of `1`.

        Returns
        -------
        lagtime : int
            The lagtime.
        """
        if self.count_model is not None:
            return self.count_model.lagtime
        return 1

    @property
    def transition_matrix(self):
        """ The transition matrix on the active set. """
        return self._transition_matrix

    @property
    def transition_matrix_tolerance(self):
        r""" The tolerance under which a matrix is considered a transition matrix. This means that all elements
        must be non-negative and the row sums must be 1. """
        return self._transition_matrix_tolerance

    def update_transition_matrix(self, value: np.ndarray):
        """ Sets the transition matrix and invalidates all cached and derived properties. """
        if value is None:
            raise ValueError("Markov state model requires a transition matrix, but it was None.")
        else:
            if not issparse(value):
                try:
                    value = np.asarray_chkfinite(value)
                except ValueError:
                    raise
            if not msmana.is_transition_matrix(value, tol=self.transition_matrix_tolerance):
                raise ValueError(f'The input transition matrix is not a stochastic matrix '
                                 f'(elements >= 0, rows sum up to 1) up to '
                                 f'tolerance {self._transition_matrix_tolerance}.')
            self._transition_matrix = value
            self._invalidate_caches()

    @cached_property
    def reversible(self) -> bool:
        """Returns whether the MarkovStateModel is reversible """
        return msmana.is_reversible(self.transition_matrix) if self._is_reversible is None else self._is_reversible

    @property
    def sparse(self) -> bool:
        """Returns whether the MarkovStateModel is sparse """
        return issparse(self.transition_matrix)

    @property
    def n_states(self) -> int:
        """ Number of active states on which all computations and estimations are done """
        return np.shape(self.transition_matrix)[0]

    @property
    def n_eigenvalues(self) -> int:
        """ number of eigenvalues to compute. """
        return self._n_eigenvalues

    @property
    def ncv(self):
        """ Number of Lanczos vectors used when computing the partial eigenvalue decomposition """
        return self._ncv

    def submodel(self, states):
        r"""
        Restricts this markov state model to a subset of states by taking a submatrix of the transition matrix
        and re-normalizing it, as well as restricting the stationary distribution and count model if given.

        Parameters
        ----------
        states : array_like of int
            states to restrict to
        Returns
        -------
        submodel : MarkovStateModel
            A onto the given states restricted MSM.
        """
        states = np.asarray(states)
        if np.any(states >= self.n_states):
            raise ValueError("At least one of the given states is not contained in this model "
                             "(n_states={}, max. given state={}).".format(self.n_states, np.max(states)))
        count_model = self.count_model
        if count_model is not None:
            count_model = count_model.submodel(states)
        transition_matrix = submatrix(self.transition_matrix, states)
        transition_matrix /= transition_matrix.sum(axis=1)[:, None]
        # set stationary distribution to None, gets recomputed in the constructor
        return MarkovStateModel(transition_matrix, stationary_distribution=None,
                                reversible=self.reversible, n_eigenvalues=min(self.n_eigenvalues, len(states)),
                                ncv=self.ncv, count_model=count_model)

    ################################################################################
    # Spectral quantities
    ################################################################################

    @cached_property
    def stationary(self):
        """ Whether the MSM is stationary, i.e. whether the initial distribution is the stationary distribution
         of the hidden transition matrix. """
        # for disconnected matrices, the stationary distribution depends on the estimator, so we can't compute
        # it directly. Therefore we test whether the initial distribution is stationary.
        return np.allclose(np.dot(self.stationary_distribution, self.transition_matrix), self.stationary_distribution)

    @cached_property
    def stationary_distribution(self):
        """The stationary distribution on the MarkovStateModel states"""
        if self._stationary_distribution is None:
            from deeptime.markov.tools.analysis import stationary_distribution as compute_sd
            stationary_distribution = compute_sd(self.transition_matrix)
            if not np.allclose(np.sum(stationary_distribution), 1., atol=1e-14):
                raise ValueError("Stationary distribution did not sum up to 1 "
                                 "(sum={})".format(np.sum(stationary_distribution)))
        else:
            stationary_distribution = self._stationary_distribution

        return stationary_distribution

    def update_stationary_distribution(self, value: Optional[np.ndarray]):
        r""" Explicitly sets the stationary distribution, re-normalizes """
        if value is not None:
            self._stationary_distribution = np.copy(value) / np.sum(value)
        else:
            self._stationary_distribution = None
        self._invalidate_caches()

    def _compute_eigenvalues(self, neig):
        """ Conducts the eigenvalue decomposition and stores k eigenvalues """
        from deeptime.markov.tools.analysis import eigenvalues as anaeig

        if self.reversible:
            self._eigenvalues = anaeig(self.transition_matrix, k=neig, ncv=self._ncv,
                                       reversible=True, mu=self.stationary_distribution)
        else:
            self._eigenvalues = anaeig(self.transition_matrix, k=neig, ncv=self._ncv, reversible=False)

        if np.all(self._eigenvalues.imag == 0):
            self._eigenvalues = self._eigenvalues.real

    def _ensure_eigenvalues(self, neig=None):
        """ Ensures that at least neig eigenvalues have been computed """
        if neig is None:
            neig = self.n_eigenvalues
        # ensure that eigenvalue decomposition with k components is done.
        try:
            m = len(self._eigenvalues)  # this will raise and exception if self._eigenvalues doesn't exist yet.
            if m < neig:
                # not enough eigenpairs present - recompute:
                self._compute_eigenvalues(neig)
        except (AttributeError, TypeError) as e:
            # no eigendecomposition yet - compute:
            self._compute_eigenvalues(neig)

    def _compute_eigendecomposition(self, n_eigenvalues: int):
        r"""
        Conducts the eigenvalue decomposition and stores k eigenvalues, left and right eigenvectors.

        Parameters
        ----------
        n_eigenvalues: int
            number of eigenvalues to compute

        Returns
        -------
        A 3-tuple consisting of the normalized right eigenvectors, a diagonal matrix with the eigenvalues, and
        the normalized left eigenvectors.

        """
        from deeptime.markov.tools.analysis import rdl_decomposition

        R, D, L = rdl_decomposition(self.transition_matrix, k=n_eigenvalues,
                                    norm='standard' if not self.reversible else 'reversible',
                                    ncv=self._ncv)
        if self.reversible:
            # everything must be real-valued
            R = R.real
            D = D.real
            L = L.real
        else:
            # if the imaginary parts are zero, discard them.
            if np.all(R.imag == 0):
                R = np.real(R)
            if np.all(D.imag == 0):
                D = np.real(D)
            if np.all(L.imag == 0):
                L = np.real(L)

        return R, D, L

    def _ensure_eigendecomposition(self, n_eigenvalues: Optional[int] = None):
        r"""
        Ensures that eigendecomposition has been performed with at least n_eigenvalues eigenpairs.
        If not, performs eigendecomposition.

        Parameters
        ----------
        n_eigenvalues : int, optional, default=None
            Number of eigenpairs required. Defaults to n_eigenvalues, see :func:`MarkovStateModel.n_eigenvalues`.
        """
        if n_eigenvalues is None:
            n_eigenvalues = self.n_eigenvalues
        # ensure that eigenvalue decomposition with k components is done.
        try:
            # raises Attribute error if this is called for the first time
            m = self._D.shape[0]
            # compute if not enough eigenpairs were computed
            compute = m < n_eigenvalues
        except AttributeError:
            compute = True
        if compute:
            self._R, self._D, self._L = self._compute_eigendecomposition(n_eigenvalues)
            self._eigenvalues = np.diag(self._D)

    def eigenvalues(self, k=None):
        r"""Compute or fetch the transition matrix eigenvalues.

        Parameters
        ----------
        k : int
            number of eigenvalues to be returned. By default will return all
            available eigenvalues

        Returns
        -------
        ts : ndarray(k,)
            transition matrix eigenvalues :math:`\lambda_i, i = 1, ..., k`.,
            sorted by descending norm.

        """
        self._ensure_eigenvalues(neig=k)
        return self._eigenvalues[:k]

    def eigenvectors_left(self, k=None):
        r"""Compute the left transition matrix eigenvectors

        Parameters
        ----------
        k : int
            number of eigenvectors to be returned. By default uses value of :attr:`n_eigenvalues`.

        Returns
        -------
        L : ndarray(k,n)
            left eigenvectors in a row matrix. l_ij is the j'th component of
            the i'th left eigenvector

        """
        self._ensure_eigendecomposition(n_eigenvalues=k)
        return self._L[:k, :]

    def eigenvectors_right(self, k=None):
        r"""Compute the right transition matrix eigenvectors

        Parameters
        ----------
        k : int
            number of eigenvectors to be computed. By default uses value of :attr:`n_eigenvalues`.

        Returns
        -------
        R : ndarray(n,k)
            right eigenvectors in a column matrix. r_ij is the i'th component
            of the j'th right eigenvector

        """
        self._ensure_eigendecomposition(n_eigenvalues=k)
        return self._R[:, :k]

    def timescales(self, k=None):
        r"""
        The relaxation timescales corresponding to the eigenvalues

        Parameters
        ----------
        k : int
            number of timescales to be returned. By default all available
            eigenvalues, minus 1.

        Returns
        -------
        ts : ndarray(m)
            relaxation timescales in units of the input trajectory time step,
            defined by :math:`-\tau / ln | \lambda_i |, i = 2,...,k+1`.

        """
        if k is None:
            self._ensure_eigenvalues()
        else:
            self._ensure_eigenvalues(neig=k + 1)
        from deeptime.markov.tools.analysis import timescales_from_eigenvalues as timescales

        ts = timescales(self._eigenvalues, tau=self.lagtime)
        if k is None:
            return ts[1:]
        else:
            return ts[1:k + 1]  # exclude the stationary process

    def propagate(self, p0, k: int):
        r""" Propagates the initial distribution p0 k times

        Computes the product

        .. math::

            p_k = p_0^T P^k

        If the lag time of transition matrix :math:`P` is :math:`\tau`, this
        will provide the probability distribution at time :math:`k \tau`.

        Parameters
        ----------
        p0 : ndarray(n,)
            Initial distribution. Vector of size of the active set.

        k : int
            Number of time steps

        Returns
        ----------
        pk : ndarray(n,)
            Distribution after k steps. Vector of size of the active set.

        """
        p0 = ensure_array(p0, ndim=1, size=self.n_states)
        assert k >= 0, 'k must be a non-negative integer'

        if k == 0:  # simply return p0 normalized
            return p0 / p0.sum()

        # sparse: we most likely don't have a full eigenvalue set, so just propagate
        if self.sparse:
            pk = np.array(p0)
            for i in range(k):
                pk = self.transition_matrix.T.dot(pk)
        else:  # dense: employ eigenvalue decomposition
            self._ensure_eigendecomposition(self.n_states)
            pk = np.linalg.multi_dot([
                p0.T, self.eigenvectors_right(), np.diag(np.power(self.eigenvalues(), k)),
                self.eigenvectors_left()
            ]).real
        return pk / pk.sum()

    ################################################################################
    # Hitting problems
    ################################################################################

    def _assert_in_active(self, A):
        """
        Checks if set A is within the active set

        Parameters
        ----------
        A : int or int array
            set of states
        """
        if np.max(A) > self.n_states:
            raise ValueError('Chosen set contains states that are not included in the active set.')

    def mfpt(self, A, B):
        """Mean first passage times from set A to set B, in units of the input trajectory time step.

        Parameters
        ----------
        A : int or int array
            set of starting states
        B : int or int array
            set of target states
        """
        from deeptime.markov.tools.analysis import mfpt
        self._assert_in_active(A)
        self._assert_in_active(B)
        return self.lagtime * mfpt(self.transition_matrix, B, origin=A, mu=self.stationary_distribution)

    def committor_forward(self, A, B):
        """Forward committor (also known as p_fold or splitting probability) from set A to set B.

        Parameters
        ----------
        A : int or int array
            set of starting states
        B : int or int array
            set of target states
        """
        from deeptime.markov.tools.analysis import committor
        self._assert_in_active(A)
        self._assert_in_active(B)
        return committor(self.transition_matrix, A, B, forward=True)

    def committor_backward(self, A, B):
        """Backward committor from set A to set B

        Parameters
        ----------
        A : int or int array
            set of starting states
        B : int or int array
            set of target states
        """
        self._assert_in_active(A)
        self._assert_in_active(B)
        from deeptime.markov.tools.analysis import committor
        return committor(self.transition_matrix, A, B, forward=False, mu=self.stationary_distribution)

    def expectation(self, a: np.ndarray):
        r"""Equilibrium expectation value of a given observable.

        Parameters
        ----------
        a : (n,) ndarray
            Observable vector on the MarkovStateModel state space

        Returns
        -------
        val: float
            Equilibrium expectation value fo the given observable

        Notes
        -----
        The equilibrium expectation value of an observable :math:`a` is defined as follows

        .. math::

            \mathbb{E}_{\mu}[a] = \sum_i \pi_i a_i

        :math:`\pi=(\pi_i)` is the stationary vector of the transition matrix :math:`P`.

        """
        a = ensure_array(a, ndim=1, size=self.n_states)
        return np.dot(a, self.stationary_distribution)

    def correlation(self, a, b=None, maxtime=None, k=None, ncv=None):
        r"""Time-correlation for equilibrium experiment.

        In order to simulate a time-correlation experiment (e.g. fluorescence
        correlation spectroscopy :cite:`msm-noe2011dynamical`, dynamical neutron
        scattering :cite:`msm-lindner2013dynamic`, ...), first compute the mean values of your experimental
        observable :math:`a` by MarkovStateModel state:

        .. math::

            a_i = \frac{1}{N_i} \sum_{x_t \in S_i} f(x_t)

        where :math:`S_i` is the set of configurations belonging to MarkovStateModel state
        :math:`i` and :math:`f()` is a function that computes the experimental
        observable of interest for configuration :math:`x_t`. If a cross-correlation
        function is wanted, also apply the above computation to a second
        experimental observable :math:`b`.

        Then the accurate (i.e. without statistical error) autocorrelation
        function of :math:`f(x_t)` given the Markov model is computed by
        correlation(a), and the accurate cross-correlation function is computed
        by correlation(a,b). This is done by evaluating the equation

        .. math::

            \begin{aligned}
            acf_a(k\tau)     &= \mathbf{a}^\top \mathrm{diag}(\boldsymbol{\pi}) \mathbf{P(\tau)}^k \mathbf{a} \\
            ccf_{a,b}(k\tau) &= \mathbf{a}^\top \mathrm{diag}(\boldsymbol{\pi}) \mathbf{P(\tau)}^k \mathbf{b}
            \end{aligned}

        where :math:`acf` stands for autocorrelation function and :math:`ccf`
        stands for cross-correlation function, :math:`\mathbf{P(\tau)}` is the
        transition matrix at lag time :math:`\tau`, :math:`\boldsymbol{\pi}` is the
        equilibrium distribution of :math:`\mathbf{P}`, and :math:`k` is the time index.

        Note that instead of using this method you could generate a long
        synthetic trajectory from the MarkovStateModel and then estimating the
        time-correlation of your observable(s) directly from this trajectory.
        However, there is no reason to do this because the present method
        does that calculation without any sampling, and only in the limit of
        an infinitely long synthetic trajectory the two results will agree
        exactly. The correlation function computed by the present method still
        has statistical uncertainty from the fact that the underlying MarkovStateModel
        transition matrix has statistical uncertainty when being estimated from
        data, but there is no additional (and unnecessary) uncertainty due to
        synthetic trajectory generation.

        Parameters
        ----------
        a : (n,) ndarray
            Observable, represented as vector on state space
        maxtime : int or float
            Maximum time (in units of the input trajectory time step) until
            which the correlation function will be evaluated.
            Internally, the correlation function can only be computed in
            integer multiples of the Markov model lag time, and therefore
            the actual last time point will be computed at :math:`\mathrm{ceil}(\mathrm{maxtime} / \tau)`
            By default (None), the maxtime will be set equal to the 5 times
            the slowest relaxation time of the MarkovStateModel, because after this time
            the signal is almost constant.
        b : (n,) ndarray (optional)
            Second observable, for cross-correlations
        k : int (optional)
            Number of eigenvalues and eigenvectors to use for computation.
            This option is only relevant for sparse matrices and long times
            for which an eigenvalue decomposition will be done instead of
            using the matrix power.
        ncv : int (optional)
            Only relevant for sparse matrices and large lag times where the
            relaxation will be computed using an eigenvalue decomposition.
            The number of Lanczos vectors generated, `ncv` must be greater than k;
            it is recommended that ncv > 2*k.

        Returns
        -------
        times : ndarray (N)
            Time points (in units of the input trajectory time step) at which
            the correlation has been computed
        correlations : ndarray (N)
            Correlation values at given times

        Examples
        --------

        This example computes the autocorrelation function of a simple observable
        on a three-state Markov model and plots the result using matplotlib:

        >>> import numpy as np
        >>> import deeptime.markov.msm as msm
        >>>
        >>> P = np.array([[0.99, 0.01, 0], [0.01, 0.9, 0.09], [0, 0.1, 0.9]])
        >>> a = np.array([0.0, 0.5, 1.0])
        >>> M = msm.MarkovStateModel(P)
        >>> times, acf = M.correlation(a)
        >>>
        >>> import matplotlib.pylab as plt # doctest: +SKIP
        >>> plt.plot(times, acf)  # doctest: +SKIP
        """
        # input checking is done in low-level API
        # compute number of tau steps
        if maxtime is None:
            # by default, use five times the longest relaxation time, because then we have relaxed to equilibrium.
            maxtime = 5 * self.timescales()[0]
        steps = np.arange(int(ceil(float(maxtime) / self.lagtime)))
        # compute correlation
        from deeptime.markov.tools.analysis import correlation
        # TODO: this could be improved. If we have already done an eigenvalue decomposition, we could provide it.
        # TODO: for this, the correlation function must accept already-available eigenvalue decompositions.
        res = correlation(self.transition_matrix, a, obs2=b, times=steps, k=k, ncv=ncv)
        # return times scaled by tau
        times = steps * self.lagtime
        return times, res

    def fingerprint_correlation(self, a, b=None, k=None, ncv=None):
        r"""Dynamical fingerprint for equilibrium time-correlation experiment.

        Parameters
        ----------
        a : (n,) ndarray
            Observable, represented as vector on MarkovStateModel state space
        b : (n,) ndarray, optional
            Second observable, for cross-correlations
        k : int, optional
            Number of eigenvalues and eigenvectors to use for computation. This
            option is only relevant for sparse matrices and long times for which
            an eigenvalue decomposition will be done instead of using the matrix
            power
        ncv : int, optional
            Only relevant for sparse matrices and large lag times, where the
            relaxation will be computed using an eigenvalue decomposition.
            The number of Lanczos vectors generated, `ncv` must be greater than k;
            it is recommended that ncv > 2*k

        Returns
        -------
        timescales : (k,) ndarray
            Time-scales (in units of the input trajectory time step) of the transition matrix
        amplitudes : (k,) ndarray
            Amplitudes for the correlation experiment

        Spectral densities are commonly used in spectroscopy. Dynamical
        fingerprints are a useful representation for computational
        spectroscopy results and have been introduced in :cite:`msm-noe2011dynamical`.

        References
        ----------
        .. footbibliography::
        """
        # input checking is done in low-level API
        # TODO: this could be improved. If we have already done an eigenvalue decomposition, we could provide it.
        # TODO: for this, the correlation function must accept already-available eigenvalue decompositions.
        from deeptime.markov.tools.analysis import fingerprint_correlation as fc
        return fc(self.transition_matrix, a, obs2=b, tau=self.lagtime, k=k, ncv=ncv)

    def relaxation(self, p0, a, maxtime=None, k=None, ncv=None):
        r"""Simulates a perturbation-relaxation experiment.

        In perturbation-relaxation experiments such as temperature-jump, pH-jump, pressure jump or rapid mixing
        experiments, an ensemble of molecules is initially prepared in an off-equilibrium distribution and the
        expectation value of some experimental observable is then followed over time as the ensemble relaxes towards
        equilibrium.

        In order to simulate such an experiment, first determine the distribution of states at which the experiment is
        started, :math:`p_0` and compute the mean values of your experimental observable :math:`a`
        by MarkovStateModel state:

        .. math::

            a_i = \frac{1}{N_i} \sum_{x_t \in S_i} f(x_t)

        where :math:`S_i` is the set of configurations belonging to MarkovStateModel state :math:`i`
        and :math:`f()` is a function that computes the experimental observable of
        interest for configuration :math:`x_t`.

        Then the accurate (i.e. without statistical error) time-dependent expectation value of :math:`f(x_t)` given the
        Markov model is computed by relaxation(p0, a). This is done by evaluating the equation

        .. math::

            E_a(k\tau) = \mathbf{p_0}^{\top} \mathbf{P(\tau)}^k \mathbf{a}

        where :math:`E` stands for the expectation value that relaxes to its equilibrium value that is identical
        to expectation(a), :math:`\mathbf{P(\tau)}` is the transition matrix at lag time :math:`\tau`,
        :math:`\boldsymbol{\pi}` is the equilibrium distribution of :math:`\mathbf{P}`, and :math:`k` is the time index.

        Note that instead of using this method you could generate many synthetic trajectory from the MarkovStateModel
        with starting points drawn from the initial distribution and then estimating the
        time-dependent expectation value by an ensemble average. However, there is no reason to do this because the
        present method does that calculation without any sampling, and only in the limit of an infinitely many
        trajectories the two results will agree exactly. The relaxation function computed by the present method still
        has statistical uncertainty from the fact that the underlying MarkovStateModel transition
        matrix has statistical uncertainty when being estimated from data, but there is no additional (and unnecessary)
        uncertainty due to synthetic trajectory generation.

        Parameters
        ----------
        p0 : (n,) ndarray
            Initial distribution for a relaxation experiment
        a : (n,) ndarray
            Observable, represented as vector on state space
        maxtime : int or float, optional
            Maximum time (in units of the input trajectory time step) until which the correlation function will be
            evaluated. Internally, the correlation function can only be computed in integer multiples of the
            Markov model lag time, and therefore the actual last time point will be computed at
            :math:`\mathrm{ceil}(\mathrm{maxtime} / \tau)`.
            By default (None), the maxtime will be set equal to the 5 times the slowest relaxation time of the
            MarkovStateModel, because after this time the signal is constant.
        k : int (optional)
            Number of eigenvalues and eigenvectors to use for computation
        ncv : int (optional)
            Only relevant for sparse matrices and large lag times, where the relaxation will be computes using an
            eigenvalue decomposition.
            The number of Lanczos vectors generated, `ncv` must be greater than k; it is recommended that ncv > 2*k

        Returns
        -------
        times : ndarray (N)
            Time points (in units of the input trajectory time step) at which the relaxation has been computed
        res : ndarray
            Array of expectation value at given times

        """
        # input checking is done in low-level API
        # compute number of tau steps
        if maxtime is None:
            # by default, use five times the longest relaxation time, because then we have relaxed to equilibrium.
            maxtime = 5 * self.timescales()[0]
        kmax = int(ceil(float(maxtime) / self.lagtime))
        steps = np.array(list(range(kmax)), dtype=int)
        # compute relaxation function
        from deeptime.markov.tools.analysis import relaxation
        # TODO: this could be improved. If we have already done an eigenvalue decomposition, we could provide it.
        # TODO: for this, the correlation function must accept already-available eigenvalue decompositions.
        res = relaxation(self.transition_matrix, p0, a, times=steps, k=k)
        # return times scaled by tau
        times = steps * self.lagtime
        return times, res

    def fingerprint_relaxation(self, p0, a, k=None, ncv=None):
        r"""Dynamical fingerprint for perturbation/relaxation experiment.

        Parameters
        ----------
        p0 : (n,) ndarray
            Initial distribution for a relaxation experiment
        a : (n,) ndarray
            Observable, represented as vector on state space
        k : int, optional
            Number of eigenvalues and eigenvectors to use for computation
        ncv : int, optional
            Only relevant for sparse matrices and large lag times, where the relaxation will be computes using an
            eigenvalue decomposition. The number of Lanczos vectors generated, `ncv` must be greater than k;
            it is recommended that ncv > 2*k

        Returns
        -------
        timescales : (k,) ndarray
            Time-scales (in units of the input trajectory time step) of the transition matrix
        amplitudes : (k,) ndarray
            Amplitudes for the relaxation experiment

        Spectral densities are commonly used in spectroscopy. Dynamical fingerprints are a useful representation
        for computational spectroscopy results and have been introduced in :cite:`msm-noe2011dynamical`.
        """
        # input checking is done in low-level API
        # TODO: this could be improved. If we have already done an eigenvalue decomposition, we could provide it.
        # TODO: for this, the correlation function must accept already-available eigenvalue decompositions.
        from deeptime.markov.tools.analysis import fingerprint_relaxation as fr
        return fr(self.transition_matrix, p0, a, tau=self.lagtime, k=k, ncv=ncv)

    def pcca(self, n_metastable_sets: int) -> PCCAModel:
        r""" Runs PCCA+ :cite:`msm-roblitz2013fuzzy` to compute a metastable decomposition of
        MarkovStateModel states.

        After calling this method you can access :func:`metastable_memberships`,
        :func:`metastable_distributions`, :func:`metastable_sets` and
        :func:`metastable_assignments` of the returned object.

        Parameters
        ----------
        n_metastable_sets : int
            Number of metastable sets

        Returns
        -------
        pcca_obj : :class:`PCCAModel <deeptime.markov.PCCAModel>`
            An object containing all PCCA+ quantities.

        Notes
        -----
        If you coarse grain with PCCA+, the order of the obtained memberships
        might not be preserved.
        """
        if not self.reversible:
            raise ValueError('Cannot compute PCCA+ for non-reversible matrices. '
                             'Set reversible=True when constructing the MarkovStateModel.')
        return pcca(self.transition_matrix, n_metastable_sets, self.stationary_distribution)

    def reactive_flux(self, source_states, target_states) -> ReactiveFlux:
        r""" A->B reactive flux from transition path theory (TPT)

        The returned object can be used to extract various quantities of the flux, as well as to
        compute A -> B transition pathways, their weights, and to coarse-grain
        the flux onto sets of states.

        Parameters
        ----------
        source_states : array_like
            List of integer state labels for set A
        target_states : array_like
            List of integer state labels for set B

        Returns
        -------
        tptobj : ReactiveFlux
            An object containing the reactive A->B flux network
            and several additional quantities, such as the stationary probability,
            committors and set definitions.

        See also
        --------
        :class:`ReactiveFlux`
            Reactive Flux model
        """
        from .._reactive_flux import reactive_flux

        return reactive_flux(
            transition_matrix=self.transition_matrix,
            source_states=source_states,
            target_states=target_states,
            stationary_distribution=self.stationary_distribution,
            transition_matrix_tolerance=None  # set to None explicitly so no check is performed
        )

    def simulate(self, n_steps: int, start: Optional[int] = None, stop: Optional[Union[int, List[int]]] = None,
                 dt: int = 1, seed: int = -1):
        r"""Generates a realization of the Markov Model.

        Parameters
        ----------
        n_steps : int
            trajectory length in steps of the lag time
        start : int, optional, default = None
            starting state
        stop : int or int-array-like, optional, default = None
            stopping hidden set. If given, the trajectory will be stopped before
            N steps once a hidden state of the stop set is reached
        dt : int, default=1
            trajectory will be saved every dt time steps.
            Internally, the dt'th power of P is taken to ensure a more efficient simulation.
        seed : int, default=-1
            If non-negative, this fixes the seed used for sampling to the provided value so that
            results are reproducible.

        Returns
        -------
        (N/dt,) ndarray
            The state trajectory with length N/dt

        Examples
        --------
        >>> msm = MarkovStateModel(transition_matrix=np.array([[.7, .3], [.3, .7]]))
        >>> trajectory = msm.simulate(n_steps=15)
        >>> print(trajectory)  # doctest:+ELLIPSIS
        [...]
        """
        if seed is None:  # the extension code internally treats -1 as default initialization
            seed = -1
        from .._markov_bindings import simulation as sim
        if start is None:
            start = np.random.choice(self.n_states, p=self.stationary_distribution)
        if self.sparse:
            transition_matrix = self.transition_matrix.toarray()
        else:
            transition_matrix = self.transition_matrix
        if dt > 1:
            transition_matrix = np.linalg.matrix_power(transition_matrix, dt)
        if stop is not None and not isinstance(stop, (list, tuple, np.ndarray)):
            stop = [stop]
        return sim.trajectory(N=n_steps, start=start, P=transition_matrix, stop=stop, seed=seed)

    ################################################################################
    # For general statistics
    ################################################################################
    def compute_trajectory_weights(self, dtrajs):
        r"""Uses the MarkovStateModel to assign a probability weight to each trajectory frame.

        This is a powerful function for the calculation of arbitrary observables in the trajectories one has
        started the analysis with. The stationary probability of the MarkovStateModel will be used to reweigh all
        states. Returns a list of weight arrays, one for each trajectory, and with a number of elements equal to
        trajectory frames. Given :math:`N` trajectories of lengths :math:`T_1` to :math:`T_N`, this function
        returns corresponding weights:

        .. math::

            (w_{1,1}, ..., w_{1,T_1}), (w_{N,1}, ..., w_{N,T_N})

        that are normalized to one:

        .. math::

            \sum_{i=1}^N \sum_{t=1}^{T_i} w_{i,t} = 1

        Suppose you are interested in computing the expectation value of a function :math:`a(x)`, where :math:`x`
        are your input configurations. Use this function to compute the weights of all input configurations and
        obtain the estimated expectation by:

        .. math::

            \langle a \rangle = \sum_{i=1}^N \sum_{t=1}^{T_i} w_{i,t} a(x_{i,t})

        Or if you are interested in computing the time-lagged correlation between functions :math:`a(x)` and
        :math:`b(x)` you could do:

        .. math::

            \langle a(t) b(t+\tau) \rangle_t = \sum_{i=1}^N \sum_{t=1}^{T_i} w_{i,t} a(x_{i,t}) a(x_{i,t+\tau})


        Returns
        -------
        weights : list of ndarray
            The normalized trajectory weights. Given :math:`N` trajectories of lengths :math:`T_1` to :math:`T_N`,
            returns the corresponding weights:

            .. math::

                (w_{1,1}, ..., w_{1,T_1}), (w_{N,1}, ..., w_{N,T_N})

        """
        # compute stationary distribution, expanded to full set
        if self.count_model is None:
            raise RuntimeError("Count model was None but needs to be provided in this case.")
        dtrajs = ensure_dtraj_list(dtrajs)
        statdist_full = np.zeros(self.count_model.n_states_full)
        statdist_full[self.count_model.state_symbols] = self.stationary_distribution
        # histogram observed states
        hist = 1.0 * count_states(dtrajs)
        # simply read off stationary distribution and accumulate total weight
        W = []
        wtot = 0.0
        for dtraj in dtrajs:
            w = statdist_full[dtraj] / hist[dtraj]
            W.append(w)
            wtot += np.sum(w)
        # normalize
        for w in W:
            w /= wtot
        # done
        return W

    def compute_state_indices(self, dtrajs) -> List[np.ndarray]:
        r"""Generates a trajectory/time indices for the given list of states. If a count model is provided in this
        MSM and it does not represent the full state space, the discrete trajectories are first mapped to the
        active state space, inactive states are mapped to -1.

        Parameters
        ----------
        dtrajs : array_like or list of array_like
            Discretized trajectories.

        Returns
        -------
        state_indices : List of ndarray
            A list of arrays with trajectory/time indices for the provided discretized trajectories.
        """
        if self.count_model is not None:
            dtrajs = self.count_model.transform_discrete_trajectories_to_submodel(dtrajs)
        if all(np.all(dtraj == -1) for dtraj in dtrajs):
            raise ValueError("None of the symbols in the discrete trajectories were represented in the "
                             "Markov state model as states.")
        return compute_index_states(dtrajs)

    ################################################################################
    # HMM-based coarse graining
    ################################################################################

    def hmm(self, dtrajs, nhidden: int, return_estimator=False):
        """Estimates a hidden Markov state model as described in :cite:`msm-noe2013projected`.

        Parameters
        ----------
        dtrajs: list of int-array like
            discrete trajectories to use for estimation of the HMM.

        nhidden : int
            number of hidden (metastable) states

        return_estimator : boolean, optional
            if False only the Model is returned,
            if True both the Estimator and the Model is returned.

        Returns
        -------
        hmm : deeptime.markov.hmm.HiddenMarkovModel
            A hidden markov model.
        """
        if not self.reversible:
            raise ValueError("Can only use HMM coarse-graining if the estimate was reversible. This is due to the use "
                             "of PCCA+ as initial estimate of the hidden transition matrix.")
        if self.sparse:
            raise ValueError("Not implemented for sparse transition matrices.")
        # check if the time-scale separation is OK
        # if hmm.n_states = msm.n_states there is no problem. Otherwise, check spectral gap
        if self.n_states > nhidden:
            ts = self.timescales()
            timescale_ratios = ts[:-1] / ts[1:]
            if timescale_ratios[nhidden - 2] < 1.5:
                import warnings
                warnings.warn(f'Requested coarse-grained model with {nhidden} metastable states at lag={self.lagtime}.'
                              f' The ratio of relaxation timescales between'
                              f' {nhidden} and {nhidden+1} states is only {timescale_ratios[nhidden - 2]}'
                              f' while we recommend at least 1.5.'
                              f' It is possible that the resulting HMM is inaccurate. Handle with caution.',
                              stacklevel=2)
        # run HMM estimate
        from deeptime.markov.hmm import MaximumLikelihoodHMM, init
        init_hmm = init.discrete.metastable_from_msm(self, nhidden, reversible=self.reversible)
        est = MaximumLikelihoodHMM(init_hmm, lagtime=self.lagtime, reversible=self.reversible)
        hmm = est.fit(dtrajs).fetch_model()
        if return_estimator:
            return est, hmm
        return hmm

    def score(self, dtrajs=None, r=2, dim=None):
        r""" Scores the MSM using the dtrajs using the variational approach for Markov processes.

        Implemented according to :cite:`msm-noe2013variational` and :cite:`msm-wu2020variational`.

        Currently only implemented using dense matrices - will be slow for large state spaces.

        Parameters
        ----------
        dtrajs : list of arrays, optional, default=None
            Test data (discrete trajectories). Note that if the test data contains states which are not
            represented in this model, they are ignored.
        r : float or str, optional, default=2
            Overwrite scoring method to be used if desired. Can be any float greater or equal 1 or 'E' for VAMP-r score
            or VAMP-E score, respectively.

            Special cases :cite:`msm-noe2013variational` :cite:`msm-wu2020variational`:

            * 'VAMP1': Sum of singular values of the symmetrized transition matrix.
              If the MSM is reversible, this is equal to the sum of transition
              matrix eigenvalues, also called Rayleigh quotient :cite:`msm-mcgibbon2015variational`.
            * 'VAMP2': Sum of squared singular values of the symmetrized  transition matrix
              :cite:`msm-wu2020variational`. If the MSM is reversible, this is equal to the
              kinetic variance :cite:`msm-noe2015kinetic`.
            * 'VAMPE': Approximation error of the estimated Koopman operator with respect to the true Koopman operator
              up to an additive constant :cite:`msm-wu2020variational`.
        dim : int or None, optional, default=None
            The maximum number of eigenvalues or singular values used in the
            score. If set to None, all available eigenvalues will be used.
        """
        koopman_model = self.empirical_koopman_model
        test_model = None
        if dtrajs is not None:
            assert self.has_count_model, "Needs count model if dtrajs are provided."
            # test data
            from deeptime.markov import TransitionCountEstimator
            cm = TransitionCountEstimator(self.count_model.lagtime, "sliding", sparse=False)\
                .fit(dtrajs).fetch_model()

            # map to present active set
            common_symbols = set(self.count_model.state_symbols).intersection(cm.state_symbols)
            common_states = self.count_model.symbols_to_states(common_symbols)
            c0t_test = np.zeros((self.n_states, self.n_states), dtype=self.transition_matrix.dtype)
            common_states_cm = cm.symbols_to_states(common_symbols)
            # only set counts for common states
            c0t_test[np.ix_(common_states, common_states)] = cm.count_matrix[np.ix_(common_states_cm, common_states_cm)]
            c00_test = np.diag(c0t_test.sum(axis=1))
            ctt_test = np.diag(c0t_test.sum(axis=0))
            assert koopman_model.cov.cov_00.shape == c00_test.shape
            assert koopman_model.cov.cov_0t.shape == c0t_test.shape
            assert koopman_model.cov.cov_tt.shape == ctt_test.shape
            test_model = CovarianceModel(c00_test, c0t_test, ctt_test)
        return vamp_score(koopman_model, r, test_model, dim=dim)


class MarkovStateModelCollection(MarkovStateModel):
    r"""A collection of Markov state models. An instance of this itself behaves like a
    :class:`MSM <MarkovStateModel>`, defaulting to the first (transition matrix, stationary distribution,
    count model) triplet provided. The model can be switched by a call to :meth:`select`.

    Parameters
    ----------
    transition_matrices : list of (n, n) ndarray
        List of transition matrices. Each of them must be row-stochastic.
    stationary_distributions : list of `None` or list of (n, ) ndarray
        List of stationary distributions belonging to transition matrices.
        If list of None, it will be evaluated upon construction of the instance.
    reversible : bool or None
        Whether the transition matrices are reversible with respect to its stationary distribution.
        If None, this is determined from the transition matrix.
    count_models : List of count models or list of None
        List of count models belonging to transition matrices.
    transition_matrix_tolerance : float
        Tolerance under which a matrix is considered to be a transition matrix.

    See Also
    --------
    MarkovStateModel
    MaximumLikelihoodMSM
    """

    def __init__(self, transition_matrices: List[np.ndarray],
                 stationary_distributions: List[np.ndarray],
                 reversible: Optional[bool],
                 count_models: List[TransitionCountModel],
                 transition_matrix_tolerance: float):
        n = len(transition_matrices)
        if n == 0:
            raise ValueError("Needs at least one transition matrix!")
        if len(stationary_distributions) != n or len(count_models) != n:
            raise ValueError(f"Got {n} transition matrices but {len(stationary_distributions)} "
                             f"stationary distributions and {len(count_models)} count models. For a one-to-one "
                             f"correspondence, these must match.")
        for i in range(n):
            if not is_square_matrix(transition_matrices[i]):
                raise ValueError("Transition matrices must be square matrices!")
            n_states = transition_matrices[i].shape[0]
            if stationary_distributions[i] is not None and len(stationary_distributions[i]) != n_states:
                raise ValueError("Lengths of stationary distributions must match respective number of states.")
            if count_models[i] is not None and count_models[i].n_states != n_states:
                raise ValueError("Number of states in count models must match states described in transition matrices.")
        self._transition_matrices = transition_matrices
        self._stationary_distributions = stationary_distributions
        self._reversible = reversible
        self._count_models = count_models
        self._transition_matrix_tolerance = transition_matrix_tolerance
        self._current_model = 0
        super(MarkovStateModelCollection, self).__init__(
            transition_matrix=self._transition_matrices[self.current_model],
            stationary_distribution=self._stationary_distributions[self.current_model],
            reversible=reversible,
            count_model=count_models[self.current_model],
            transition_matrix_tolerance=transition_matrix_tolerance
        )

    @property
    def current_model(self) -> int:
        r""" The currently selected model index.

        :type: int
        """
        return self._current_model

    @property
    def n_connected_msms(self) -> int:
        r"""Number of markov state models in this collection.

        :type: int
        """
        return len(self._transition_matrices)

    def state_symbols(self, model_index: Optional[int] = None) -> np.ndarray:
        r""" Yields the state symbols of a particular model in this collection. Can only be called if the corresponding
        model has a count model.

        Parameters
        ----------
        model_index : int, optional, default=None
            The model index. If None, this evaluates to :meth:`current_model`.

        Returns
        -------
        symbols : (n,) ndarray
            The state symbols associated with the model index. If model index is None, it is defaulted to the
            current model.
        """
        if model_index is None:
            model_index = self.current_model
        self._validate_model_index(model_index)
        if self._count_models[model_index] is None:
            raise ValueError("This can only be evaluated if count models are provided at construction time.")
        return self._count_models[model_index].state_symbols

    def _validate_model_index(self, model_index):
        r"""Validates whether the model index is valid (ie not out of bounds). If it is in bounds, this is a no-op,
        otherwise raises an IndexError.

        Parameters
        ----------
        model_index : int
            The model index

        Raises
        ------
        IndexError
            If the model index is out of bounds.
        """
        if model_index >= self.n_connected_msms:
            raise IndexError(f"There are only {self.n_connected_msms} MSMs in this collection, but "
                             f"information about MSM {model_index} was requested.")

    @property
    def state_fractions(self):
        r""" Yields the fractions of states represented in each of the models in this collection. """
        return [counts.selected_state_fraction for counts in self._count_models]

    @property
    def state_fraction(self):
        r""" The fraction of states represented in the selected model. """
        return self.state_fractions[self.current_model]

    @property
    def count_fractions(self):
        r""" Yields the fraction of counts represented in each of the models in this collection. Calling this method
        assumes that the MSMs in the collection stem from actual data with state statistics embedded in the count
        models. """
        return [counts.selected_count_fraction if counts is not None else None for counts in self._count_models]

    @property
    def count_fraction(self):
        r""" The fraction of counts represented in the selected model. """
        return self.count_fractions[self.current_model]

    def select(self, model_index):
        r""" Selects a different model in the collection. Changes the behavior of the collection to mimic a MSM
        associated to respective transition matrix, stationary distribution, and count model.

        Parameters
        ----------
        model_index : int
            The model index.

        Raises
        ------
        IndexError
            If model index is out of bounds.

        """
        self._validate_model_index(model_index)
        self._invalidate_caches()
        super().__init__(transition_matrix=self._transition_matrices[model_index],
                         stationary_distribution=self._stationary_distributions[model_index],
                         reversible=self.reversible, count_model=self._count_models[model_index],
                         transition_matrix_tolerance=self.transition_matrix_tolerance)
        self._current_model = model_index


def _svd_sym_koopman(msm: MarkovStateModel, empirical: bool, epsilon=1e-10) -> CovarianceKoopmanModel:
    """ Computes the SVD of the symmetrized Koopman operator in the empirical distribution, returns as Koopman model.
    """
    K = msm.transition_matrix
    if empirical:
        C = msm.count_model.count_matrix
        if issparse(C):
            C = C.toarray()
        cov0t = C
        cov00 = np.diag(cov0t.sum(axis=1))
        covtt = np.diag(cov0t.sum(axis=0))
        cov = CovarianceModel(cov_00=cov00, cov_0t=cov0t, cov_tt=covtt)
        # reweight operator to empirical distribution
        C0t_re = cov00 @ K
        # symmetrized operator and SVD
        K_sym = np.linalg.multi_dot([spd_inv_sqrt(cov.cov_00, epsilon=epsilon), C0t_re,
                                     spd_inv_sqrt(cov.cov_tt, epsilon=epsilon)])
    else:
        cov00 = np.diag(msm.stationary_distribution)
        covtt = np.diag(msm.stationary_distribution)
        cov0t = cov00 @ K
        cov = CovarianceModel(cov_00=cov00, cov_0t=cov0t, cov_tt=covtt)
        K_sym = np.linalg.multi_dot([spd_inv_sqrt(cov.cov_00, epsilon=epsilon), cov0t,
                                     spd_inv_sqrt(cov.cov_tt, epsilon=epsilon)])
    U, S, Vt = scipy.linalg.svd(K_sym, compute_uv=True)
    U = spd_inv_sqrt(cov.cov_00, epsilon=epsilon) @ U
    Vt = Vt @ spd_inv_sqrt(cov.cov_tt, epsilon=epsilon)
    return CovarianceKoopmanModel(U, S, Vt.T, cov, K.shape[0], K.shape[0])



from scipy.sparse import issparse


def submatrix(M, sel):
    """Returns a submatrix of the quadratic matrix M, given by the selected columns and row
    Parameters
    ----------
    M : ndarray(n,n)
        symmetric matrix
    sel : int-array
        selection of rows and columns. Element i,j will be selected if both are in sel.
    Returns
    -------
    S : ndarray(m,m)
        submatrix with m=len(sel)
    """
    assert len(M.shape) == 2, 'M is not a matrix'
    assert M.shape[0] == M.shape[1], 'M is not quadratic'
    """Row slicing"""
    if issparse(M):
        C_cc = M.tocsr()
    else:
        C_cc = M
    C_cc = C_cc[sel, :]

    """Column slicing"""
    if issparse(M):
        C_cc = C_cc.tocsc()
    C_cc = C_cc[:, sel]

    if issparse(M):
        return C_cc.tocoo()

    return C_cc



import numbers
import warnings

import numpy as np

from ._moments import moments_XX, moments_XXXY, moments_block

__author__ = 'noe'


class Moments:

    def __init__(self, w, sx, sy, Mxy):
        r"""
        Parameters
        ----------
        w : float
            statistical weight.
                w = \sum_t w_t
            In most cases, :math:`w_t=1`, and then w is just the number of samples that went into s1, S2.
        s : ndarray(n,)
            sum over samples:
            .. math:
                s = \sum_t w_t x_t
        M : ndarray(n, n)
            .. math:
                M = (X-s)^T (X-s)
        """
        self.w = float(w)
        self.sx = sx
        self.sy = sy
        self.Mxy = Mxy

    def copy(self):
        return Moments(self.w, self.sx.copy(), self.sy.copy(), self.Mxy.copy())

    def combine(self, other, mean_free=False):
        """ According to `this report <http://i.stanford.edu/pub/cstr/reports/cs/tr/79/773/CS-TR-79-773.pdf>`__.
        """
        w1 = self.w
        w2 = other.w
        w = w1 + w2
        q = w2 / w1
        dsx = q * self.sx - other.sx
        dsy = q * self.sy - other.sy
        # update
        self.w = w1 + w2
        self.sx = self.sx + other.sx
        self.sy = self.sy + other.sy
        #
        if mean_free:
            if len(self.Mxy.shape) == 1:  # diagonal only
                d = dsx*dsy
            else:
                d = np.outer(dsx, dsy)
            self.Mxy += other.Mxy + (w1 / (w2 * w)) * d
        else:
            self.Mxy += other.Mxy
        return self

    @property
    def mean_x(self):
        return self.sx / self.w

    @property
    def mean_y(self):
        return self.sy / self.w

    def covar(self, bessels_correction=True):
        """
        Return covariance matrix:

        Parameters:
        -----------
        bessels_correction : bool, optional, default=True
            Use Bessel's correction in order to
            obtain an unbiased estimator of sample covariances.

        """
        if bessels_correction:
            return self.Mxy / (self.w - 1)
        else:
            return self.Mxy / self.w


class MomentsStorage:

    def __init__(self, nsave, remove_mean=False, rtol=1.5):
        """
        Parameters
        ----------
        rtol : float
            To decide when to merge two Moments. Ideally I'd like to merge two
            Moments when they have equal weights (i.e. equally many data points
            went into them). If we always add data chunks with equal weights,
            this can be achieved by using a binary tree, i.e. let M1 be the
            moment estimates from one chunk. Two of them are added to M2, Two
            M2 are added to M4, and so on. This way you need to store log2
            (n_chunks) number of Moment estimates.
            In practice you might get data in chunks of unequal length or weight.
            Therefore we need some heuristic when two Moment estimates should get
            merged. This is the role of rtol.

        """
        self.nsave = nsave
        self.storage = []
        self.rtol = rtol
        self.remove_mean = remove_mean

    def _can_merge_tail(self):
        """ Checks if the two last list elements can be merged
        """
        if len(self.storage) < 2:
            return False
        return self.storage[-2].w <= self.storage[-1].w * self.rtol

    def store(self, moments):
        """ Store object X with weight w
        """
        if len(self.storage) == self.nsave:  # merge if we must
            self.storage[-1].combine(moments, mean_free=self.remove_mean)
        else:  # append otherwise
            self.storage.append(moments)
        # merge if possible
        while self._can_merge_tail():
            M = self.storage.pop()
            self.storage[-1].combine(M, mean_free=self.remove_mean)

    @property
    def moments(self):
        """
        """
        # collapse storage if necessary
        while len(self.storage) > 1:
            M = self.storage.pop()
            self.storage[-1].combine(M, mean_free=self.remove_mean)
        return self.storage[0]

    def clear(self):
        self.storage.clear()


class RunningCovar:
    """ Running covariance estimator

    Estimator object that can be fed chunks of X and Y data, and
    that can generate on-the-fly estimates of mean, covariance, running sum
    and second moment matrix.

    Parameters
    ----------
    compute_XX : bool
        Estimate the covariance of X
    compute_XY : bool
        Estimate the cross-covariance of X and Y
    compute_YY : bool
        Estimate the covariance of Y
    remove_mean : bool
        Remove the data mean in the covariance estimation
    symmetrize : bool
        Use symmetric estimates with sum defined by sum_t x_t + y_t and
        second moment matrices defined by X'X + Y'Y and Y'X + X'Y.
    modify_data : bool
        If remove_mean=True, the mean will be removed in the input data,
        without creating an independent copy. This option is faster but should
        only be selected if the input data is not used elsewhere.
    sparse_mode : str
        one of:
            * 'dense' : always use dense mode
            * 'sparse' : always use sparse mode if possible
            * 'auto' : automatic
    diag_only: bool
        If True, the computation is restricted to the diagonal entries (autocorrelations) only.
    nsave : int
        Depth of Moment storage. Moments computed from each chunk will be
        combined with Moments of similar statistical weight using the pairwise
        combination algorithm described in :footcite:`chan1982updating`.

    References
    ----------
    .. footbibliography::
    """

    # to get the Y mean, but this is currently not stored.
    def __init__(self, compute_XX=True, compute_XY=False, compute_YY=False,
                 remove_mean=False, symmetrize=False, sparse_mode='auto', modify_data=False,
                 diag_only=False, nsave=5):
        # check input
        if not compute_XX and not compute_XY:
            raise ValueError('One of compute_XX or compute_XY must be True.')
        if symmetrize and compute_YY:
            raise ValueError('Combining compute_YY and symmetrize=True is meaningless.')
        if symmetrize and not compute_XY:
            warnings.warn('symmetrize=True has no effect with compute_XY=False.')
        if diag_only and sparse_mode != 'dense':
            if sparse_mode == 'sparse':
                warnings.warn('Computing diagonal entries only is not implemented for sparse mode. '
                              'Switching to dense mode.')
            sparse_mode = 'dense'
        # storage
        self.compute_XX = compute_XX
        self.compute_XY = compute_XY
        self.compute_YY = compute_YY

        self.storage_XX = MomentsStorage(nsave, remove_mean=remove_mean)
        self.storage_XY = MomentsStorage(nsave, remove_mean=remove_mean)
        self.storage_YY = MomentsStorage(nsave, remove_mean=remove_mean)
        # symmetry
        self.remove_mean = remove_mean
        self.symmetrize = symmetrize
        # flags
        self.sparse_mode = sparse_mode
        self.modify_data = modify_data
        # whether to compute only matrix diagonals
        self.diag_only = diag_only

    def add(self, X, Y=None, weights=None, column_selection=None):
        """
        Add trajectory to estimate.

        Parameters
        ----------
        X : ndarray(T, N)
            array of N time series.
        Y : ndarray(T, N)
            array of N time series, usually time shifted version of X.
        weights : None or float or ndarray(T, ):
            weights assigned to each trajectory point. If None, all data points have weight one. If float,
            the same weight will be given to all data points. If ndarray, each data point is assigned a separate
            weight.
        column_selection : ndarray((d,), dtype=int), optional, default=None
            Optional column subselection.
        """

        # check input
        X = X if X.ndim >= 2 else X[..., np.newaxis]
        if Y is not None:
            Y = Y if Y.ndim >= 2 else Y[..., np.newaxis]
        T = X.shape[0]
        if Y is not None:
            assert Y.shape[0] == T, 'X and Y must have equal length'
        if column_selection is not None and self.diag_only:
            raise ValueError('Computing only parts of the diagonal is not supported.')

        # Weights cannot be used for compute_YY:
        if weights is not None and self.compute_YY:
            raise ValueError('Use of weights is not implemented for compute_YY==True')
        if weights is not None:
            # Convert to array of length T if weights is a single number:
            if isinstance(weights, numbers.Real):
                weights = weights * np.ones(T, dtype=float)
            # Check appropriate length if weights is an array:
            elif isinstance(weights, np.ndarray):
                if len(weights) != T:
                    raise ValueError(f'Weights and X must have equal length. '
                                     f'Was {len(weights)} and {len(X)}, respectively.')
            else:
                raise TypeError(f'Weights is of type {type(weights)}, must be a number or ndarray.')
        # estimate and add to storage
        if self.compute_XX and not self.compute_XY and not self.compute_YY:
            w, s_X, C_XX = moments_XX(X, remove_mean=self.remove_mean, weights=weights, sparse_mode=self.sparse_mode,
                                      modify_data=self.modify_data, column_selection=column_selection,
                                      diag_only=self.diag_only)
            if column_selection is not None:
                s_Xk = s_X[column_selection]
            else:
                s_Xk = s_X
            self.storage_XX.store(Moments(w, s_X, s_Xk, C_XX))
        elif self.compute_XX and self.compute_XY and not self.compute_YY:
            assert Y is not None
            w, s_X, s_Y, C_XX, C_XY = moments_XXXY(X, Y, remove_mean=self.remove_mean, symmetrize=self.symmetrize,
                                                   weights=weights, sparse_mode=self.sparse_mode, modify_data=self.modify_data,
                                                   column_selection=column_selection, diag_only=self.diag_only)
            # make copy in order to get independently mergeable moments
            if column_selection is not None:
                s_Xk = s_X[column_selection]
                s_Yk = s_Y[column_selection]
            else:
                s_Xk = s_X
                s_Yk = s_Y
            self.storage_XX.store(Moments(w, s_X, s_Xk, C_XX))
            self.storage_XY.store(Moments(w, s_X, s_Yk, C_XY))
        else:  # compute block
            assert Y is not None
            assert not self.symmetrize
            w, s, C = moments_block(X, Y, remove_mean=self.remove_mean,
                                    sparse_mode=self.sparse_mode, modify_data=self.modify_data,
                                    column_selection=column_selection, diag_only=self.diag_only)
            # make copy in order to get independently mergeable moments
            if column_selection is not None:
                s0k = s[0][column_selection]
                s1k = s[1][column_selection]
            else:
                s0k = s[0]
                s1k = s[1]
            if self.compute_XX:
                self.storage_XX.store(Moments(w, s[0], s0k, C[0][0]))
            if self.compute_XY:
                self.storage_XY.store(Moments(w, s[0], s1k, C[0][1]))
            self.storage_YY.store(Moments(w, s[1], s1k, C[1][1]))

    def sum_X(self):
        if self.compute_XX:
            return self.storage_XX.moments.sx
        elif self.compute_XY:
            return self.storage_XY.moments.sx
        else:
            raise RuntimeError('sum_X is not available')

    def sum_Y(self):
        if self.compute_XY:
            return self.storage_XY.moments.sy
        elif self.compute_YY:
            return self.storage_YY.moments.sy
        else:
            raise RuntimeError('sum_Y is not available')

    def mean_X(self):
        if self.compute_XX:
            return self.storage_XX.moments.mean_x
        elif self.compute_XY:
            return self.storage_XY.moments.mean_y
        else:
            raise RuntimeError('mean_X is not available')

    def mean_Y(self):
        if self.compute_XY:
            return self.storage_XY.moments.mean_y
        elif self.compute_YY:
            return self.storage_YY.moments.mean_y
        else:
            raise RuntimeError('mean_Y is not available')

    def weight_XX(self):
        return self.storage_XX.moments.w

    def weight_XY(self):
        return self.storage_XY.moments.w

    def weight_YY(self):
        return self.storage_YY.moments.w

    def moments_XX(self):
        return self.storage_XX.moments.Mxy

    def moments_XY(self):
        return self.storage_XY.moments.Mxy

    def moments_YY(self):
        return self.storage_YY.moments.Mxy

    def cov_XX(self, bessel=True):
        return self.storage_XX.moments.covar(bessels_correction=bessel)

    def cov_XY(self, bessel=True):
        return self.storage_XY.moments.covar(bessels_correction=bessel)

    def cov_YY(self, bessel):
        return self.storage_YY.moments.covar(bessels_correction=bessel)

    def clear(self):
        self.storage_XX.clear()
        self.storage_XY.clear()
        self.storage_YY.clear()


def running_covar(xx=True, xy=False, yy=False, remove_mean=False, symmetrize=False, sparse_mode='auto',
                  modify_data=False, diag_only=False, nsave=5):
    """ Returns a running covariance estimator

    Returns an estimator object that can be fed chunks of X and Y data, and
    that can generate on-the-fly estimates of mean, covariance, running sum
    and second moment matrix.

    Parameters
    ----------
    xx : bool
        Estimate the covariance of X
    xy : bool
        Estimate the cross-covariance of X and Y
    yy : bool
        Estimate the covariance of Y
    remove_mean : bool
        Remove the data mean in the covariance estimation
    symmetrize : bool
        Use symmetric estimates with sum defined by sum_t x_t + y_t and
        second moment matrices defined by X'X + Y'Y and Y'X + X'Y.
    modify_data : bool
        If remove_mean=True, the mean will be removed in the input data,
        without creating an independent copy. This option is faster but should
        only be selected if the input data is not used elsewhere.
    sparse_mode : str
        one of:
            * 'dense' : always use dense mode
            * 'sparse' : always use sparse mode if possible
            * 'auto' : automatic
    diag_only: bool
        If True, the computation is restricted to the diagonal entries (autocorrelations) only.
    nsave : int
        Depth of Moment storage. Moments computed from each chunk will be
        combined with Moments of similar statistical weight using the pairwise
        combination algorithm described in :footcite:`chan1982updating`.

    References
    ----------
    .. footbibliography::
    """
    return RunningCovar(compute_XX=xx, compute_XY=xy, compute_YY=yy, sparse_mode=sparse_mode, modify_data=modify_data,
                        remove_mean=remove_mean, symmetrize=symmetrize,
                        diag_only=diag_only, nsave=nsave)



from typing import Optional, List

import numpy as np

from ..base import Dataset


def timeshifted_split(inputs, lagtime: int, chunksize: int = 1000, stride: int = 1, n_splits: Optional[int] = None,
                      shuffle: bool = False, random_state: Optional[np.random.RandomState] = None):
    r""" Utility function which splits input trajectories into pairs of timeshifted data :math:`(X_t, X_{t+\tau})`.
    In case multiple trajectories are provided, the timeshifted pairs are always within the same trajectory.

    Parameters
    ----------
    inputs : (T, n) ndarray or list of (T_i, n) ndarrays
        Input trajectory or trajectories. In case multiple trajectories are provided, they must have the same dimension
        in the second axis but may be of variable length.
    lagtime : int
        The lag time :math:`\tau` used to produce timeshifted blocks.
    chunksize : int, default=1000
        The chunk size, i.e., the maximal length of the blocks.
    stride: int, default=1
        Optional stride which is applied *after* creating a tau-shifted version of the dataset.
    n_splits : int, optional, default=None
        Alternative to chunksize - this determines the number of timeshifted blocks that is drawn from each provided
        trajectory. Supersedes whatever was provided as chunksize.
    shuffle : bool, default=False
        Whether to shuffle the data prior to splitting it.
    random_state : np.random.RandomState, default=None
        When shuffling this can be used to set a specific random state.

    Returns
    -------
    iterable : Generator
        A Python generator which can be iterated.

    Examples
    --------
    Using chunksize:

    >>> data = np.array([0, 1, 2, 3, 4, 5, 6])
    >>> for X, Y in timeshifted_split(data, lagtime=1, chunksize=4):
    ...     print(X, Y)
    [0 1 2 3] [1 2 3 4]
    [4 5] [5 6]

    Using n_splits:

    >>> data = np.array([0, 1, 2, 3, 4, 5, 6])
    >>> for X, Y in timeshifted_split(data, lagtime=1, n_splits=2):
    ...     print(X, Y)
    [0 1 2] [1 2 3]
    [3 4 5] [4 5 6]
    """
    if lagtime < 0:
        raise ValueError('lagtime has to be non-negative')
    if int(chunksize) < 0:
        raise ValueError('chunksize has to be positive')

    if shuffle and random_state is None:
        random_state = np.random.RandomState()

    if not isinstance(inputs, list):
        inputs = [inputs]

    if not all(len(data) > lagtime for data in inputs):
        too_short_inputs = [i for i, x in enumerate(inputs) if len(x) <= lagtime]
        raise ValueError(f'Input contained to short (smaller than lagtime({lagtime}) at following '
                         f'indices: {too_short_inputs}')

    for data in inputs:
        data = np.asarray_chkfinite(data)
        data_lagged = data[lagtime:][::stride]
        if lagtime > 0:
            data = data[:-lagtime][::stride]
        else:
            data = data[0::stride]  # otherwise data is empty as slice over `data[:-0]`
        ix = np.arange(len(data))  # iota range over data
        if shuffle:
            random_state.shuffle(ix)

        if n_splits is not None:
            assert n_splits >= 1
            for ix_split in np.array_split(ix, n_splits):
                if len(ix_split) > 0:
                    x = data[ix_split]
                    if lagtime > 0:
                        x_lagged = data_lagged[ix_split]
                        yield x, x_lagged
                    else:
                        yield x
                else:
                    break
        else:
            t = 0
            while t < len(data):
                if t == len(data_lagged):
                    break
                x = data[ix[t:min(t + chunksize, len(data))]]
                if lagtime > 0:
                    x_lagged = data_lagged[ix[t:min(t + chunksize, len(data_lagged))]]
                    yield x, x_lagged
                else:
                    yield x
                t += chunksize


class ConcatDataset(Dataset):
    r""" Concatenates existing datasets.

    Parameters
    ----------
    datasets : list of dataset
        Datasets to concatenate.
    """

    def __init__(self, datasets: List[Dataset]):
        self._lengths = [len(ds) for ds in datasets]
        self._cumlen = np.cumsum(self._lengths)
        self._datasets = datasets

    def setflags(self, write=True):
        for ds in self._datasets:
            ds.setflags(write=write)

    @property
    def subsets(self):
        r""" Returns the list of datasets this concat dataset is composed of.

        :type: list of dataset
        """
        return self._datasets

    def _dataset_index(self, ix):
        from bisect import bisect_right
        ds_index = bisect_right(self._cumlen, ix)
        item_index = ix if ds_index == 0 else ix - self._cumlen[ds_index - 1]
        return ds_index, item_index

    def __getitem__(self, ix):
        ds_index, item_index = self._dataset_index(ix)
        return self._datasets[ds_index][item_index]

    def __len__(self):
        return self._cumlen[-1]


class TimeLaggedDataset(Dataset):
    r""" High-level container for time-lagged time-series data.
    This can be used together with pytorch data tools, i.e., data loaders and other utilities.

    Parameters
    ----------
    data : (T, n) ndarray
        The data which is wrapped into a dataset
    data_lagged : (T, m) ndarray
        Corresponding timelagged data. Must be of same length.

    See Also
    --------
    TimeLaggedConcatDataset, TrajectoryDataset, TrajectoriesDataset
    """

    def __init__(self, data, data_lagged):
        assert len(data) == len(data_lagged), \
            f"Length of trajectory for data and data_lagged does not match ({len(data)} != {len(data_lagged)})"
        self._data = data
        self._data_lagged = data_lagged

    def setflags(self, write=True):
        self._data.setflags(write=write)
        self._data_lagged.setflags(write=write)

    def astype(self, dtype):
        r""" Sets the datatype of contained arrays and returns a new instance of TimeLaggedDataset.

        Parameters
        ----------
        dtype
            The new dtype.

        Returns
        -------
        converted_ds : TimeLaggedDataset
            The dataset with converted dtype.
        """
        return TimeLaggedDataset(self._data.astype(dtype), self._data_lagged.astype(dtype))

    @property
    def data(self) -> np.ndarray:
        r""" Instantaneous data. """
        return self._data

    @property
    def data_lagged(self) -> np.ndarray:
        r""" Time-lagged data. """
        return self._data_lagged

    def __getitem__(self, item):
        return self._data[item], self._data_lagged[item]

    def __len__(self):
        return len(self._data)


class TimeLaggedConcatDataset(ConcatDataset):
    r""" Specialization of the :class:`ConcatDataset` which uses that all subsets are time lagged datasets, offering
    fancy and more efficient slicing / getting items.

    Parameters
    ----------
    datasets : list of TimeLaggedDataset
        The input datasets
    """

    def __init__(self, datasets: List[TimeLaggedDataset]):
        assert all(isinstance(x, TimeLaggedDataset) for x in datasets)
        super().__init__(datasets)

    @staticmethod
    def _compute_overlap(stride, traj_len, skip):
        r""" Given two trajectories :math:`T_1` and :math:`T_2`, this function calculates for the first trajectory
        an overlap, i.e., a skip parameter for :math:`T_2` such that the trajectory fragments
        :math:`T_1` and :math:`T_2` appear as one under the given stride.

        :param stride: the (global) stride parameter
        :param traj_len: length of T_1
        :param skip: skip of T_1
        :return: skip of T_2

        Notes
        -----
        Idea for deriving the formula: It is

        .. code::
            K = ((traj_len - skip - 1) // stride + 1) = #(data points in trajectory of length (traj_len - skip)).

        Therefore, the first point's position that is not contained in :math:`T_1` anymore is given by

        .. code::
            pos = skip + s * K.

        Thus the needed skip of :math:`T_2` such that the same stride parameter makes :math:`T_1` and :math:`T_2`
        "look as one" is

        .. code::
            overlap = pos - traj_len.
        """
        return stride * ((traj_len - skip - 1) // stride + 1) - traj_len + skip

    def __getitem__(self, ix):
        if isinstance(ix, slice):
            xs, ys = [], []
            end_ds, end_ix = self._dataset_index(ix.stop if ix.stop is not None else len(self))
            start_ds, start = self._dataset_index(ix.start if ix.start is not None else 0)
            stride = ix.step if ix.step is not None else 1
            for ds in range(start_ds, end_ds + 1):
                stop_ix = self._lengths[ds] if ds != end_ds else end_ix

                if stop_ix > start and ds < len(self._lengths):
                    local_slice = slice(start, stop_ix, stride)
                    xs.append(self._datasets[ds].data[local_slice])
                    ys.append(self._datasets[ds].data_lagged[local_slice])
                    start = self._compute_overlap(stride, self._lengths[ds], start)
            return np.concatenate(xs), np.concatenate(ys)
        else:
            return super().__getitem__(ix)


class TrajectoryDataset(TimeLaggedDataset):
    r"""Creates a trajectory dataset from a single trajectory by applying a lagtime.

    Parameters
    ----------
    lagtime : int
        Lagtime, must be positive. The effective size of the dataset reduces by the selected lagtime.
    trajectory : (T, d) ndarray
        Trajectory with T frames in d dimensions.

    Raises
    ------
    AssertionError
        If lagtime is not positive or trajectory is too short for lagtime.
    """

    def __init__(self, lagtime, trajectory):
        assert lagtime > 0, "Lagtime must be positive"
        assert len(trajectory) > lagtime, "Not enough data to satisfy lagtime"
        super().__init__(trajectory[:-lagtime], trajectory[lagtime:])
        self._trajectory = trajectory
        self._lagtime = lagtime

    @property
    def lagtime(self):
        return self._lagtime

    @property
    def trajectory(self):
        return self._trajectory

    @staticmethod
    def from_trajectories(lagtime, data: List[np.ndarray]):
        r""" Creates a time series dataset from multiples trajectories by applying a lagtime.

        Parameters
        ----------
        lagtime : int
            Lagtime, must be positive. The effective size of the dataset reduces by the selected lagtime.
        data : list of ndarray
            List of trajectories.

        Returns
        -------
        dataset : TrajectoriesDataset
            Concatenation of timeseries datasets.

        Raises
        ------
        AssertionError
            If data is empty, lagtime is not positive,
            the shapes do not match, or lagtime is too long for any of the trajectories.
        """
        return TrajectoriesDataset.from_numpy(lagtime, data)


class TrajectoriesDataset(TimeLaggedConcatDataset):
    r""" Dataset composed of multiple trajectories.

    Parameters
    ----------
    data : list of TrajectoryDataset
        The trajectories in form of trajectory datasets.

    See Also
    --------
    TrajectoryDataset.from_trajectories
        Method to create a TrajectoriesDataset from multiple raw data trajectories.
    """

    def __init__(self, data: List[TrajectoryDataset]):
        assert len(data) > 0, "List of data should not be empty."
        assert all(x.lagtime == data[0].lagtime for x in data), "Lagtime must agree"
        super().__init__(data)

    @staticmethod
    def from_numpy(lagtime, data: List[np.ndarray]):
        r""" Creates a time series dataset from multiples trajectories by applying a lagtime.

        Parameters
        ----------
        lagtime : int
            Lagtime, must be positive. The effective size of the dataset reduces by the selected lagtime.
        data : list of ndarray
            List of trajectories.

        Returns
        -------
        dataset : TrajectoriesDataset
            Concatenation of timeseries datasets.

        Raises
        ------
        AssertionError
            If data is empty, lagtime is not positive,
            the shapes do not match, or lagtime is too long for any of the trajectories.
        """
        assert len(data) > 0 and all(data[0].shape[1:] == x.shape[1:] for x in data), "Shape mismatch!"
        return TrajectoriesDataset([TrajectoryDataset(lagtime, traj) for traj in data])

    @property
    def lagtime(self):
        r""" The lagtime.

        :type: int
        """
        return self.subsets[0].lagtime

    @property
    def trajectories(self):
        r""" Contained raw trajectories.

        :type: list of ndarray
        """
        return [x.trajectory for x in self.subsets]



import numpy as np


class SamplerRevPi:

    def __init__(self, C, pi, P0=None, P_mle=None, eps=0.1, seed=-1):
        from deeptime.markov.tools.estimation.dense.mle import mle_trev_given_pi
        from .._mle_bindings import RevPiSampler32, RevPiSampler64, RevPiSampler128

        dtype = C.dtype
        if dtype not in (np.float32, np.float64, np.longdouble):
            dtype = np.float64
        self.C = C.astype(dtype)
        self.pi = pi.astype(dtype)

        if P_mle is None:
            P_mle = mle_trev_given_pi(C, pi)

        if P0 is None:
            cdiag = np.diag(C)
            """Entries with cii=0"""
            ind = (cdiag == 0)
            """Add counts, s.t. cii+bii>0 for all i"""
            bdiag = np.zeros_like(cdiag)
            bdiag[ind] = 1.0
            B = np.diag(bdiag)
            P0 = mle_trev_given_pi(C + B, pi)

        """Diagonal prior parameters"""
        b = np.zeros(C.shape[0])

        cii = C[np.diag_indices(C.shape[0])]

        """Zero diagonal entries of C"""
        ind1 = np.isclose(cii, 0.0)
        b[ind1] = eps

        """Non-zero diagonal entries of P0"""
        pii0 = P_mle[np.diag_indices(P_mle.shape[0])]
        ind2 = (pii0 > 0.0)

        """Find elements pii0>0 and cii=0"""
        ind3 = np.logical_and(ind1, ind2)
        b[ind3] = 1.0

        self.b = b

        """Initial state of the chain"""
        self.X = pi[:, np.newaxis] * P0
        self.X /= self.X.sum()

        """Check for valid input"""
        self.check_input()

        """Set up index arrays"""
        self.I, self.J = np.where((self.C + self.C.T) > 0.0)

        """Init the xsampler"""
        if self.C.dtype == np.float32:
            self.xsampler = RevPiSampler32(seed)
            self.X = self.X.astype(np.float32)
            self.b = self.b.astype(np.float32)
            self.pi = self.pi.astype(np.float32)
        elif self.C.dtype == np.float64:
            self.xsampler = RevPiSampler64(seed)
            self.X = self.X.astype(np.float64)
            self.b = self.b.astype(np.float64)
            self.pi = self.pi.astype(np.float64)
        elif self.C.dtype == np.longdouble:
            self.xsampler = RevPiSampler128(seed)
            self.X = self.X.astype(np.longdouble)
            self.b = self.b.astype(np.longdouble)
            self.pi = self.pi.astype(np.longdouble)
        else:
            raise ValueError(f"Unknown dtype {self.C.dtype}")

    def check_input(self):
        from deeptime.markov.tools.analysis import is_connected

        if not np.all(self.C >= 0):
            raise ValueError("Count matrix contains negative elements")
        if not is_connected(self.C, directed=False):
            raise ValueError("Count matrix is not connected")
        if not np.all(self.X >= 0.0):
            raise ValueError("P0 contains negative entries")
        if not np.allclose(self.X, self.X.T):
            raise ValueError("P0 is not reversible")
        """Check sparsity pattern - ignore diagonal"""
        C_sym = self.C + self.C.T
        X_sym = self.X + self.X.T
        ind = np.diag_indices(C_sym.shape[0])
        C_sym[ind] = 0.0
        X_sym[ind] = 0.0
        iC, jC = np.where(C_sym > 0.0)
        iV, jV = np.where(X_sym > 0.0)
        if not np.array_equal(iC, iV):
            raise ValueError('Sparsity patterns of C and X are different.')
        if not np.array_equal(jC, jV):
            raise ValueError('Sparsity patterns of C and X are different.')

    def update(self, N=1):
        for i in range(N):
            self.xsampler.update(self.C, self.X, self.b)

    def sample(self, N=1, return_statdist=False):
        self.update(N=N)
        P = self.X / self.pi[:, np.newaxis]
        if return_statdist:
            return P, self.pi
        else:
            return P




def module_available(modname: str) -> bool:
    r"""Checks whether a module is installed and available for import by the current interpreter.

    Parameters
    ----------
    modname : str
        Name of the module

    Returns
    -------
    available: bool
        Whether the module is available.
    """
    try:
        __import__(modname)
        return True
    except ImportError:
        return False


def handle_progress_bar(progress):
    r"""Takes a (potential) progress bar, if None, just returns an iterable that does nothing but return everything.

    Parameters
    ----------
    progress : progress bar or None, optional
        The progressbar

    Returns
    -------
    progress_bar : iterable
        A progress bar (or no/identity progress bar if input was None).
    """
    if progress is None:
        def progress(iterable, *args, **kw):
            for x in iterable:
                yield x
    return progress



from typing import Optional, List

import numpy as np

import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from ...base import Transformer, Model, EstimatorTransformer
from ...base_torch import DLEstimatorMixin
from ...util.torch import map_data, MLP


class TAEModel(Model, Transformer):
    r""" Model produced by time-lagged autoencoders. Contains the encoder, decoder, and can transform data to
    the latent code.

    Parameters
    ----------
    encoder : torch.nn.Module
        The encoder module.
    decoder : torch.nn.Module
        The decoder module.
    device : torch device or None, default=None
        The device to use.
    dtype : numpy datatype, default=np.float32
        The dtype that is used for transformation of data.

    See Also
    --------
    TAE
    """

    def __init__(self, encoder, decoder, device=None, dtype=np.float32):
        self._encoder = encoder
        self._decoder = decoder
        self._device = device
        self._dtype = dtype

    @property
    def encoder(self):
        r""" The encoder.

        :type: torch.nn.Module
        """
        return self._encoder

    @property
    def decoder(self):
        r""" The decoder.

        :type: torch.nn.Module
        """
        return self._decoder

    def _encode(self, x: torch.Tensor):
        return self._encoder(x)

    def transform(self, data, **kwargs):
        r""" Transforms a trajectory (or a list of trajectories) by passing them through the encoder network.

        Parameters
        ----------
        data : array_like or list of array_like
            The trajectory data.
        **kwargs
            Ignored.

        Returns
        -------
        latent_code : ndarray or list of ndarray
            The trajectory / trajectories encoded to the latent representation.
        """
        out = []
        for data_tensor in map_data(data, device=self._device, dtype=self._dtype):
            out.append(self._encode(data_tensor).cpu().numpy())
        return out if len(out) > 1 else out[0]


class TAE(EstimatorTransformer, DLEstimatorMixin):
    r""" Time-lagged autoencoder. :footcite:`wehmeyer2018timelagged`

    Parameters
    ----------
    encoder : torch.nn.Module
        Encoder module.
    decoder : torch.nn.Module
        Decoder module, its input features should be compatible with the encoder's output features.
    optimizer : str or callable, default='Adam'
        The optimizer to use, defaults to Adam. If given as string, can be one of 'Adam', 'SGD', 'RMSProp'.
        In case of a callable, the callable should take a `params` parameter list and a `lr` learning rate, yielding
        an optimizer instance based on that.
    learning_rate : float, default=3e-4
        The learning rate that is used for the optimizer. Defaults to the Karpathy learning rate.

    References
    ----------
    .. footbibliography::
    """
    _MUTABLE_INPUT_DATA = True

    def __init__(self, encoder: nn.Module, decoder: nn.Module, optimizer='Adam', learning_rate=3e-4):
        super().__init__()
        self._encoder = encoder
        self._decoder = decoder
        self.learning_rate = learning_rate
        self.setup_optimizer(optimizer, list(encoder.parameters()) + list(decoder.parameters()))
        self._mse_loss = nn.MSELoss(reduction='sum')
        self._train_losses = []
        self._val_losses = []

    def evaluate_loss(self, x: torch.Tensor, y: torch.Tensor):
        r""" Evaluates the loss based on input tensors.

        Parameters
        ----------
        x : torch.Tensor
            The tensor that is passed through encoder and decoder networks.
        y : torch.Tensor
            The tensor the forward pass is compared against.

        Returns
        -------
        loss : torch.Tensor
            The loss.
        """
        return self._mse_loss(y, self._decoder(self._encoder(x)))

    @property
    def train_losses(self) -> np.ndarray:
        r""" The collected train scores. First dimension contains the step, second dimension the score. Initially empty.

        :type: (T, 2) ndarray
        """
        return np.array(self._train_losses)

    @property
    def validation_losses(self) -> np.ndarray:
        r""" The collected validation scores. First dimension contains the step, second dimension the score.
        Initially empty.

        :type: (T, 2) ndarray
        """
        return np.array(self._val_losses)

    def fit(self, data_loader: DataLoader, n_epochs: int = 5, validation_loader: Optional[DataLoader] = None, **kwargs):
        r""" Fits the encoder and decoder based on data. Note that a call to fit does not reset the weights in the
        networks that are currently in :attr:`encoder` and :attr:`decoder`.

        Parameters
        ----------
        data_loader : DataLoader
            Data loader which yields batches of instantaneous and time-lagged data.
        n_epochs : int, default=5
            Number of epochs to train for.
        validation_loader : DataLoader, optional, default=MNone
            Data loader which  yields batches of instantaneous and time-lagged data for validation purposes. Can be
            left None, in which case no validation is performed.
        **kwargs
            Ignored kw.

        Returns
        -------
        self : TAE
            Reference to self.
        """
        step = 0
        for epoch in range(n_epochs):

            self._encoder.train()
            self._decoder.train()
            for batch_0, batch_t in data_loader:
                batch_0 = batch_0.to(device=self.device)
                batch_t = batch_t.to(device=self.device)

                self.optimizer.zero_grad()
                loss_value = self.evaluate_loss(batch_0, batch_t)
                loss_value.backward()
                self.optimizer.step()

                self._train_losses.append((step, loss_value.item()))

                step += 1

            if validation_loader is not None:
                self._encoder.eval()
                self._decoder.eval()

                with torch.no_grad():
                    lval = []
                    for batch_0, batch_t in data_loader:
                        batch_0 = batch_0.to(device=self.device)
                        batch_t = batch_t.to(device=self.device)

                        loss_value = self.evaluate_loss(batch_0, batch_t).item()
                        lval.append(loss_value)
                    self._val_losses.append((step, np.mean(lval)))

        return self

    def fetch_model(self) -> TAEModel:
        r""" Yields a new instance of :class:`TAEModel`.

        .. warning::
            The model can be subject to side-effects in case :meth:`fit` is called multiple times, as no deep copy
            is performed of encoder and decoder networks.

        Returns
        -------
        model : TAEModel
            The model.
        """
        from copy import deepcopy
        return TAEModel(deepcopy(self._encoder), deepcopy(self._decoder), device=self.device, dtype=self.dtype)


def _reparameterize(mu: torch.Tensor, logvar: torch.Tensor) -> torch.Tensor:
    std = torch.exp(0.5 * logvar)
    eps = torch.randn_like(std)
    return eps * std + mu


class TVAEModel(TAEModel):
    r""" Model produced by the time-lagged variational autoencoder (:class:`TVAE`).
    When transforming data, the encoded mean and log-variance are reparametrized and yielded.

    See Also
    --------
    TAEModel
    """
    def _encode(self, x: torch.Tensor):
        return _reparameterize(*self.encoder(x))


class TVAEEncoder(MLP):
    r""" A kind of :class:`MLP`, which maps the output through not one but two transformations so that it projects
    onto mean and log-variance.

    Parameters
    ----------
    units : list of int
        The units of the network.
    nonlinearity : callable
        The nonlinearity to use. Callable must produce a `torch.nn.Module` which implements the nonlinear operation.
    """

    def __init__(self, units: List[int], nonlinearity=nn.ELU):
        super().__init__(units[:-1], nonlinearity=nonlinearity, initial_batchnorm=False,
                         output_nonlinearity=nonlinearity)
        lat_in = units[-2]
        lat_out = units[-1]
        self._to_mu = nn.Linear(lat_in, lat_out)
        self._to_logvar = nn.Linear(lat_in, lat_out)

    def forward(self, inputs):
        out = self._sequential(inputs)
        return self._to_mu(out), self._to_logvar(out)


class TVAE(TAE):
    r""" The time-lagged variational autoencoder. For a description of the arguments please see :class:`TAE`.
    The one additional argument is `beta`, which is the KLD-weight during optimization.

    See Also
    --------
    TAE
    TVAEModel
    """
    def __init__(self, encoder: nn.Module, decoder: nn.Module, optimizer='Adam', learning_rate: float = 5e-4,
                 beta: float = 1.):
        super().__init__(encoder, decoder, optimizer=optimizer, learning_rate=learning_rate)
        self._beta = beta

    def evaluate_loss(self, x: torch.Tensor, y: torch.Tensor):
        r""" Evaluates the reconstruction loss and latent regularization loss, returns the sum.

        Parameters
        ----------
        x : torch.nn.Tensor
            Input tensor to be passed through the network.
        y : torch.nn.Tensor
            Tensor which `x` is compared against.

        Returns
        -------
        loss : torch.nn.Tensor
            The loss.
        """
        mu, logvar = self._encoder(x)
        z = _reparameterize(mu, logvar)
        y_hat = self._decoder(z)
        mse_val = self._mse_loss(y_hat, y)
        kld = torch.mean(-0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp(), dim=1), dim=0)
        return mse_val + self._beta * kld / float(y.shape[1])

    def fetch_model(self) -> TVAEModel:
        r""" Yields a new instance of :class:`TVAEModel`.

        .. warning::
            The model can be subject to side-effects in case :meth:`fit` is called multiple times, as no deep copy
            is performed of encoder and decoder networks.

        Returns
        -------
        model : TVAEModel
            The model.
        """
        from copy import deepcopy
        return TVAEModel(deepcopy(self._encoder), deepcopy(self._decoder), self.device, self.dtype)



"""
Created on 22.11.2013

@author: Jan-Hendrik Prinz
"""

import numpy
from .._stationary_vector import stationary_distribution


# TODO:make faster. So far not effectively programmed
# Martin: done, but untested, since there is no testcase...
def forward_committor_sensitivity(T, A, B, index):
    """
    calculate the sensitivity matrix for index of the forward committor from A to B given transition matrix T.
    Parameters
    ----------
    T : numpy.ndarray shape = (n, n)
        Transition matrix
    A : array like
        List of integer state labels for set A
    B : array like
        List of integer state labels for set B
    index : entry of the committor for which the sensitivity is to be computed

    Returns
    -------
    x : ndarray, shape=(n, n)
        Sensitivity matrix for entry index around transition matrix T. Reversibility is not assumed.
    """

    n = len(T)
    set_X = numpy.arange(n)  # set(range(n))
    set_A = numpy.unique(A)  # set(A)
    set_B = numpy.unique(B)  # set(B)
    set_AB = numpy.union1d(set_A, set_B)  # set_A | set_B
    notAB = numpy.setdiff1d(set_X, set_AB, True)  # list(set_X - set_AB)
    m = len(notAB)

    K = T - numpy.diag(numpy.ones(n))

    U = K[numpy.ix_(notAB.tolist(), notAB.tolist())]

    v = numpy.zeros(m)

    # for i in xrange(0, m):
    #   for k in xrange(0, len(set_B)):
    #       v[i] = v[i] - K[notAB[i], B[k]]
    v[:] = v[:] - K[notAB[:], B[:]]

    qI = numpy.linalg.solve(U, v)

    q_forward = numpy.zeros(n)
    #q_forward[set_A] = 0 # double assignment.
    q_forward[set_B] = 1
    #for i in range(len(notAB)):
    q_forward[notAB[:]] = qI[:]

    target = numpy.eye(1, n, index)
    target = target[0, notAB]

    UinvVec = numpy.linalg.solve(U.T, target)
    Siab = numpy.zeros((n, n))

    for i in range(m):
        Siab[notAB[i]] = - UinvVec[i] * q_forward

    return Siab


def backward_committor_sensitivity(T, A, B, index):
    """
    calculate the sensitivity matrix for index of the backward committor from A to B given transition matrix T.
    Parameters
    ----------
    T : numpy.ndarray shape = (n, n)
        Transition matrix
    A : array like
        List of integer state labels for set A
    B : array like
        List of integer state labels for set B
    index : entry of the committor for which the sensitivity is to be computed

    Returns
    -------
    x : ndarray, shape=(n, n)
        Sensitivity matrix for entry index around transition matrix T. Reversibility is not assumed.
    """

    # This is really ugly to compute. The problem is, that changes in T induce changes in
    # the stationary distribution and so we need to add this influence, too
    # I implemented something which is correct, but don't ask me about the derivation

    n = len(T)

    trT = numpy.transpose(T)

    one = numpy.ones(n)
    eq = stationary_distribution(T)

    mEQ = numpy.diag(eq)
    mIEQ = numpy.diag(1.0 / eq)
    mSEQ = numpy.diag(1.0 / eq / eq)

    backT = numpy.dot(mIEQ, numpy.dot(trT, mEQ))

    qMat = forward_committor_sensitivity(backT, A, B, index)

    matA = trT - numpy.identity(n)
    matA = numpy.concatenate((matA, [one]))

    phiM = numpy.linalg.pinv(matA)

    phiM = phiM[:, 0:n]

    trQMat = numpy.transpose(qMat)

    d1 = numpy.dot(mSEQ, numpy.diagonal(numpy.dot(numpy.dot(trT, mEQ), trQMat), 0))
    d2 = numpy.diagonal(numpy.dot(numpy.dot(trQMat, mIEQ), trT), 0)

    psi1 = numpy.dot(d1, phiM)
    psi2 = numpy.dot(-d2, phiM)

    v1 = psi1 - one * numpy.dot(psi1, eq)
    v3 = psi2 - one * numpy.dot(psi2, eq)

    part1 = numpy.outer(eq, v1)
    part2 = numpy.dot(numpy.dot(mEQ, trQMat), mIEQ)
    part3 = numpy.outer(eq, v3)

    sensitivity = part1 + part2 + part3

    return sensitivity


def eigenvalue_sensitivity(T, k):
    """
    calculate the sensitivity matrix for eigenvalue k given transition matrix T.
    Parameters
    ----------
    T : numpy.ndarray shape = (n, n)
        Transition matrix
    k : int
        eigenvalue index for eigenvalues order descending

    Returns
    -------
    x : ndarray, shape=(n, n)
        Sensitivity matrix for entry index around transition matrix T. Reversibility is not assumed.
    """

    eValues, rightEigenvectors = numpy.linalg.eig(T)
    leftEigenvectors = numpy.linalg.inv(rightEigenvectors)

    perm = numpy.argsort(eValues)[::-1]

    rightEigenvectors = rightEigenvectors[:, perm]
    leftEigenvectors = leftEigenvectors[perm]

    sensitivity = numpy.outer(leftEigenvectors[k], rightEigenvectors[:, k])

    return sensitivity


def timescale_sensitivity(T, k):
    """
    calculate the sensitivity matrix for timescale k given transition matrix T.
    Parameters
    ----------
    T : numpy.ndarray shape = (n, n)
        Transition matrix
    k : int
        timescale index for timescales of descending order (k = 0 for the infinite one)

    Returns
    -------
    x : ndarray, shape=(n, n)
        Sensitivity matrix for entry index around transition matrix T. Reversibility is not assumed.
    """

    eValues, rightEigenvectors = numpy.linalg.eig(T)
    leftEigenvectors = numpy.linalg.inv(rightEigenvectors)

    perm = numpy.argsort(eValues)[::-1]

    eValues = eValues[perm]
    rightEigenvectors = rightEigenvectors[:, perm]
    leftEigenvectors = leftEigenvectors[perm]

    eVal = eValues[k]

    sensitivity = numpy.outer(leftEigenvectors[k], rightEigenvectors[:, k])

    if eVal < 1.0:
        factor = 1.0 / (numpy.log(eVal) ** 2) / eVal
    else:
        factor = 0.0

    sensitivity *= factor

    return sensitivity


# TODO: The eigenvector sensitivity depends on the normalization, e.g. l^T r = 1 or norm(r) = 1
# Should we fix that or add another option. Also the sensitivity depends on the initial eigenvectors
# Now everything is set to use norm(v) = 1 for left and right
# In the case of the stationary distribution we want sum(pi) = 1, so this function
# does NOT return the same as stationary_distribution_sensitivity if we choose k = 0 and right = False!

# TODO: If we choose k = 0 and right = False we might throw a warning!?!

def eigenvector_sensitivity(T, k, j, right=True):
    """
    calculate the sensitivity matrix for entry j of left or right eigenvector k given transition matrix T.
    Parameters
    ----------
    T : numpy.ndarray shape = (n, n)
        Transition matrix
    k : int
        eigenvector index ordered with descending eigenvalues
    j : int
        entry of eigenvector k for which the sensitivity is to be computed
    right : boolean (default: True)
        If set to True (default) the right eigenvectors are considered, otherwise the left ones

    Returns
    -------
    x : ndarray, shape=(n, n)
        Sensitivity matrix for entry index around transition matrix T. Reversibility is not assumed.

    Remarks
    -------
    Eigenvectors can naturally be scaled and so will their sensitivity depend on their size.
    For that reason we need to agree on a normalization for which the sensitivity is computed.
    Here we use the natural norm(vector) = 1 condition which is different from the results, e.g.
    the rdl_decomposition returns.
    This is especially important for the stationary distribution, which is the first left eigenvector.
    For this reason this function return a different sensitivity for the first left eigenvector
    than the function stationary_distribution_sensitivity and this function should not be used in this
    case!
    """

    n = len(T)

    if not right:
        T = numpy.transpose(T)

    eValues, rightEigenvectors = numpy.linalg.eig(T)
    leftEigenvectors = numpy.linalg.inv(rightEigenvectors)
    perm = numpy.argsort(eValues)[::-1]

    eValues = eValues[perm]
    rightEigenvectors = rightEigenvectors[:, perm]
    leftEigenvectors = leftEigenvectors[perm]

    rEV = rightEigenvectors[:, k]
    lEV = leftEigenvectors[k]
    eVal = eValues[k]

    vecA = numpy.zeros(n)
    vecA[j] = 1.0

    matA = T - eVal * numpy.identity(n)
    # Use here rEV as additional condition, means that we assume the vector to be
    # orthogonal to rEV
    matA = numpy.concatenate((matA, [rEV]))

    phi = numpy.linalg.lstsq(numpy.transpose(matA), vecA, rcond=-1)

    phi = numpy.delete(phi[0], -1)

    sensitivity = -numpy.outer(phi, rEV) + numpy.dot(phi, rEV) * numpy.outer(lEV, rEV)

    if not right:
        sensitivity = numpy.transpose(sensitivity)

    return sensitivity


def stationary_distribution_sensitivity(T, j):
    r"""Calculate the sensitivity matrix for entry j the stationary
    distribution vector given transition matrix T.

    Parameters
    ----------
    T : numpy.ndarray shape = (n, n)
        Transition matrix
    j : int
        entry of stationary distribution for which the sensitivity is to be computed

    Returns
    -------
    x : ndarray, shape=(n, n)
        Sensitivity matrix for entry index around transition matrix T. Reversibility is not assumed.

    Remark
    ------
    Note, that this function uses a different normalization convention for the sensitivity compared to
    eigenvector_sensitivity. See there for further information.
    """

    n = len(T)

    lEV = numpy.ones(n)
    rEV = stationary_distribution(T)
    eVal = 1.0

    T = numpy.transpose(T)

    vecA = numpy.zeros(n)
    vecA[j] = 1.0

    matA = T - eVal * numpy.identity(n)
    # normalize s.t. sum is one using rEV which is constant
    matA = numpy.concatenate((matA, [lEV]))

    phi = numpy.linalg.lstsq(numpy.transpose(matA), vecA, rcond=-1)
    phi = numpy.delete(phi[0], -1)

    sensitivity = -numpy.outer(rEV, phi) + numpy.dot(phi, rEV) * numpy.outer(rEV, lEV)

    return sensitivity


def mfpt_sensitivity(T, target, j):
    """
    calculate the sensitivity matrix for entry j of the mean first passage time (MFPT) given transition matrix T.
    Parameters
    ----------
    T : numpy.ndarray shape = (n, n)
        Transition matrix
    target : int
        target state to which the MFPT is computed
    j : int
        entry of the mfpt vector for which the sensitivity is to be computed

    Returns
    -------
    x : ndarray, shape=(n, n)
        Sensitivity matrix for entry index around transition matrix T. Reversibility is not assumed.
    """

    n = len(T)

    matA = T - numpy.diag(numpy.ones((n)))
    matA[target] *= 0
    matA[target, target] = 1.0

    tVec = -1. * numpy.ones(n)
    tVec[target] = 0

    mfpt = numpy.linalg.solve(matA, tVec)
    aVec = numpy.zeros(n)
    aVec[j] = 1.0

    phiVec = numpy.linalg.solve(numpy.transpose(matA), aVec)

    # TODO: Check sign of sensitivity!

    sensitivity = -1.0 * numpy.outer(phiVec, mfpt)
    sensitivity[target] *= 0

    return sensitivity


def expectation_sensitivity(T, a):
    r"""Sensitivity of expectation value of observable A=(a_i).

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix
    a : (M,) ndarray
        Observable, a[i] is the value of the observable at state i.

    Returns
    -------
    S : (M, M) ndarray
        Sensitivity matrix of the expectation value.

    """
    M = T.shape[0]
    S = numpy.zeros((M, M))
    for i in range(M):
        S += a[i] * stationary_distribution_sensitivity(T, i)
    return S



import numpy as np
from scipy.linalg import solve
from scipy.sparse.linalg import spsolve

import scipy.sparse as sparse

from ._stationary_vector import stationary_distribution


def _set_up_linear_system(K, A, B):
    """Assemble left-hand side W for linear system"""
    """Equation (I)"""
    W = 1.0 * K
    """Equation (II)"""
    if sparse.issparse(W):
        W = W.todok()
        W[list(A), :] = 0.0
        W.tocsr()
        W = W + sparse.coo_matrix((np.ones(len(A)), (list(A), list(A))), shape=W.shape).tocsr()
    else:
        W[list(A), :] = 0.0
        W[list(A), list(A)] = 1.0
    """Equation (III)"""
    if sparse.issparse(W):
        W = W.todok()
        W[list(B), :] = 0.0
        W.tocsr()
        W = W + sparse.coo_matrix((np.ones(len(B)), (list(B), list(B))), shape=W.shape).tocsr()
    else:
        W[list(B), :] = 0.0
        W[list(B), list(B)] = 1.0
    return W


def forward_committor(T, A, B):
    r"""Forward committor between given sets.

    The forward committor u(x) between sets A and B is the probability
    for the chain starting in x to reach B before reaching A.

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix
    A : array_like
        List of integer state labels for set A
    B : array_like
        List of integer state labels for set B

    Returns
    -------
    u : (M, ) ndarray
        Vector of forward committor probabilities

    Notes
    -----
    The forward committor is a solution to the following
    boundary-value problem

    .. math::

        \sum_j L_{ij} u_{j}=0    for i in X\(A u B) (I)
                      u_{i}=0    for i \in A        (II)
                      u_{i}=1    for i \in B        (III)

    with generator matrix L=(P-I).

    """
    X = set(range(T.shape[0]))
    A = set(A)
    B = set(B)
    AB = A.intersection(B)
    if len(AB) > 0:
        raise ValueError("Sets A and B have to be disjoint")
    L = T - np.eye(T.shape[0])  # Generator matrix

    W = _set_up_linear_system(L, A, B)
    """Assemble right hand side r for linear system"""
    """Equation (I+II)"""
    r = np.zeros(T.shape[0])
    """Equation (III)"""
    r[list(B)] = 1.0

    u = solve(W, r) if not sparse.issparse(W) else spsolve(W, r)
    return u


def backward_committor(T, A, B, mu=None):
    r"""Backward committor between given sets.

    The backward committor u(x) between sets A and B is the
    probability for the chain starting in x to have come from A last
    rather than from B.

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix
    A : array_like
        List of integer state labels for set A
    B : array_like
        List of integer state labels for set B
    mu : (M, ) ndarray (optional)
        Stationary vector

    Returns
    -------
    u : (M, ) ndarray
        Vector of forward committor probabilities

    Notes
    -----
    The forward committor is a solution to the following
    boundary-value problem

    .. math::

        \sum_j K_{ij} \pi_{j} u_{j}=0    for i in X\(A u B) (I)
                                  u_{i}=1    for i \in A        (II)
                                  u_{i}=0    for i \in B        (III)

    with adjoint of the generator matrix K=(D_pi(P-I))'.

    """
    X = set(range(T.shape[0]))
    A = set(A)
    B = set(B)
    AB = A.intersection(B)
    if len(AB) > 0:
        raise ValueError("Sets A and B have to be disjoint")
    if mu is None:
        mu = stationary_distribution(T)
    if sparse.issparse(T):
        L = T - sparse.eye(T.shape[0], T.shape[0])
        D = sparse.diags([mu, ], [0, ])
        K = (D.dot(L)).T
    else:
        K = np.transpose(mu[:, np.newaxis] * (T - np.eye(T.shape[0])))

    """Assemble left-hand side W for linear system"""
    W = _set_up_linear_system(K, A, B)
    """Assemble right-hand side r for linear system"""
    """Equation (I)+(III)"""
    r = np.zeros(T.shape[0])
    """Equation (II)"""
    r[list(A)] = 1.0

    u = solve(W, r) if not sparse.issparse(W) else spsolve(W, r)

    return u



"""
Created on 29.11.2013

.. moduleauthor:: marscher, noe
"""

import numpy as np

from .._decomposition import rdl_decomposition


def time_correlation_by_diagonalization(P, pi, obs1, obs2=None, time=1, rdl=None):
    """
    calculates time correlation. Raises P to power 'times' by diagonalization.
    If rdl tuple (R, D, L) is given, it will be used for
    further calculation.
    """
    if rdl is None:
        raise ValueError("no rdl decomposition")
    R, D, L = rdl

    d_times = np.diag(D) ** time
    diag_inds = np.diag_indices_from(D)
    D_time = np.zeros(D.shape, dtype=d_times.dtype)
    D_time[diag_inds] = d_times
    P_time = np.dot(np.dot(R, D_time), L)

    # multiply element-wise obs1 and pi. this is obs1' diag(pi)
    l = np.multiply(obs1, pi)
    m = np.dot(P_time, obs2)
    result = np.dot(l, m)
    return result


def time_correlation_direct_by_mtx_vec_prod(P, mu, obs1, obs2=None, time=1, start_values=None, return_P_k_obs=False):
    r"""Compute time-correlation of obs1, or time-cross-correlation with obs2.

    The time-correlation at time=k is computed by the matrix-vector expression:
    cor(k) = obs1' diag(pi) P^k obs2


    Parameters
    ----------
    P : ndarray, shape=(n, n) or scipy.sparse matrix
        Transition matrix
    obs1 : ndarray, shape=(n)
        Vector representing observable 1 on discrete states
    obs2 : ndarray, shape=(n)
        Vector representing observable 2 on discrete states. If not given,
        the autocorrelation of obs1 will be computed
    mu : ndarray, shape=(n)
        stationary distribution vector.
    time : int
        time point at which the (auto)correlation will be evaluated.
    start_values : (time, ndarray <P, <P, obs2>>_t)
        start iteration of calculation of matrix power product, with this values.
        only useful when calling this function out of a loop over times.
    return_P_k_obs : bool
        if True, the dot product <P^time, obs2> will be returned for further
        calculations.

    Returns
    -------
    cor(k) : float
           correlation between observations
    """
    # input checks
    if not (type(time) == int):
        if not (type(time) == np.int64):
            raise TypeError("given time (%s) is not an integer, but has type: %s"
                            % (str(time), type(time)))
    if obs1.shape[0] != P.shape[0]:
        raise ValueError("observable shape not compatible with given matrix")
    if obs2 is None:
        obs2 = obs1
    # multiply element-wise obs1 and pi. this is obs1' diag(pi)
    l = np.multiply(obs1, mu)
    # raise transition matrix to power of time by substituting dot product
    # <Pk, obs2> with something like <P, <P, obs2>>.
    # This saves a lot of matrix matrix multiplications.
    if start_values:  # begin with a previous calculated val
        P_i_obs = start_values[1]
        # calculate difference properly!
        time_prev = start_values[0]
        t_diff = time - time_prev
        r = range(t_diff)
    else:
        if time >= 2:
            P_i_obs = np.dot(P, np.dot(P, obs2))  # vector <P, <P, obs2> := P^2 * obs
            r = range(time - 2)
        elif time == 1:
            P_i_obs = np.dot(P, obs2)  # P^1 = P*obs
            r = range(0)
        elif time == 0:  # P^0 = I => I*obs2 = obs2
            P_i_obs = obs2
            r = range(0)

    for k in r:  # since we already substituted started with 0
        P_i_obs = np.dot(P, P_i_obs)
    corr = np.dot(l, P_i_obs)
    if return_P_k_obs:
        return corr, (time, P_i_obs)
    else:
        return corr


def time_correlations_direct(P, pi, obs1, obs2=None, times=[1]):
    r"""Compute time-correlations of obs1, or time-cross-correlation with obs2.

    The time-correlation at time=k is computed by the matrix-vector expression:
    cor(k) = obs1' diag(pi) P^k obs2


    Parameters
    ----------
    P : ndarray, shape=(n, n) or scipy.sparse matrix
        Transition matrix
    obs1 : ndarray, shape=(n)
        Vector representing observable 1 on discrete states
    obs2 : ndarray, shape=(n)
        Vector representing observable 2 on discrete states. If not given,
        the autocorrelation of obs1 will be computed
    pi : ndarray, shape=(n)
        stationary distribution vector. Will be computed if not given
    times : array-like, shape(n_t)
        Vector of time points at which the (auto)correlation will be evaluated

    Returns
    -------

    """
    n_t = len(times)
    times = np.sort(times)  # sort it to use caching of previously computed correlations
    f = np.zeros(n_t)

    # maximum time > number of rows?
    if times[-1] > P.shape[0]:
        use_diagonalization = True
        R, D, L = rdl_decomposition(P)
        # discard imaginary part, if all elements i=0
        if not np.any(np.iscomplex(R)):
            R = np.real(R)
        if not np.any(np.iscomplex(D)):
            D = np.real(D)
        if not np.any(np.iscomplex(L)):
            L = np.real(L)
        rdl = (R, D, L)

    if use_diagonalization:
        for i in range(n_t):
            f[i] = time_correlation_by_diagonalization(P, pi, obs1, obs2, times[i], rdl)
    else:
        start_values = None
        for i in range(n_t):
            f[i], start_values = \
                time_correlation_direct_by_mtx_vec_prod(P, pi, obs1, obs2,
                                                        times[i], start_values, True)
    return f


def time_relaxation_direct_by_mtx_vec_prod(P, p0, obs, time=1, start_values=None, return_pP_k=False):
    r"""Compute time-relaxations of obs with respect of given initial distribution.

    relaxation(k) = p0 P^k obs

    Parameters
    ----------
    P : ndarray, shape=(n, n) or scipy.sparse matrix
        Transition matrix
    p0 : ndarray, shape=(n)
        initial distribution
    obs : ndarray, shape=(n)
        Vector representing observable on discrete states.
    time : int or array like
        time point at which the (auto)correlation will be evaluated.

    start_values = (time,
    Returns
    -------
    relaxation : float
    """
    # input checks
    if not type(time) == int:
        if not type(time) == np.int64:
            raise TypeError("given time (%s) is not an integer, but has type: %s"
                            % (str(time), type(time)))
    if obs.shape[0] != P.shape[0]:
        raise ValueError("observable shape not compatible with given matrix")
    if p0.shape[0] != P.shape[0]:
        raise ValueError("shape of init dist p0 (%s) not compatible with given matrix (shape=%s)"
                         % (p0.shape[0], P.shape))
    # propagate in time
    if start_values:  # begin with a previous calculated val
        pk_i = start_values[1]
        time_prev = start_values[0]
        t_diff = time - time_prev
        r = range(t_diff)
    else:
        if time >= 2:
            pk_i = np.dot(np.dot(p0, P), P)  # pk_2
            r = range(time - 2)
        elif time == 1:
            pk_i = np.dot(p0, P)  # propagate once
            r = range(0)
        elif time == 0:  # P^0 = I => p0*I = p0
            pk_i = p0
            r = range(0)

    for k in r:  # perform the rest of the propagations p0 P^t_diff
        pk_i = np.dot(pk_i, P)

    # result
    l = np.dot(pk_i, obs)
    if return_pP_k:
        return l, (time, pk_i)
    else:
        return l


def time_relaxation_direct_by_diagonalization(P, p0, obs, time, rdl=None):
    if rdl is None:
        raise ValueError("no rdl decomposition")
    R, D, L = rdl

    d_times = np.diag(D) ** time
    diag_inds = np.diag_indices_from(D)
    D_time = np.zeros(D.shape, dtype=d_times.dtype)
    D_time[diag_inds] = d_times
    P_time = np.dot(np.dot(R, D_time), L)

    result = np.dot(np.dot(p0, P_time), obs)
    return result


def time_relaxations_direct(P, p0, obs, times=[1]):
    r"""Compute time-relaxations of obs with respect of given initial distribution.

    relaxation(k) = p0 P^k obs

    Parameters
    ----------
    P : ndarray, shape=(n, n) or scipy.sparse matrix
        Transition matrix
    p0 : ndarray, shape=(n)
        initial distribution
    obs : ndarray, shape=(n)
        Vector representing observable on discrete states.
    times : array-like, shape(n_t)
        Vector of time points at which the (auto)correlation will be evaluated

    Returns
    -------
    relaxations : ndarray, shape(n_t)
    """
    n_t = len(times)
    times = np.sort(times)

    # maximum time > number of rows?
    if times[-1] > P.shape[0]:
        use_diagonalization = True
        R, D, L = rdl_decomposition(P)
        # discard imaginary part, if all elements i=0
        if not np.any(np.iscomplex(R)):
            R = np.real(R)
        if not np.any(np.iscomplex(D)):
            D = np.real(D)
        if not np.any(np.iscomplex(L)):
            L = np.real(L)
        rdl = (R, D, L)

    f = np.empty(n_t, dtype=D.dtype)

    if use_diagonalization:
        for i in range(n_t):
            f[i] = time_relaxation_direct_by_diagonalization(P, p0, obs, times[i], rdl)
    else:
        start_values = None
        for i in range(n_t):
            f[i], start_values = time_relaxation_direct_by_mtx_vec_prod(P, p0, obs, times[i], start_values, True)
    return f



def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('markov', parent_package, top_path)

    config.add_extension('_markov_bindings',
                         sources=['_bindings/src/markov_module.cpp'],
                         include_dirs=['_bindings/include'],
                         language='c++',
                         )

    config.add_subpackage('msm')
    config.add_subpackage('hmm')
    config.add_subpackage('tools')
    config.add_subpackage('sample')

    return config



r"""Transition matrix sampling module. Provides a common class for sampling of

i) non-reversible transition matrices
ii) reversible transition matrices
iii) reversible transition matrices with fixed stationary vector

from given data

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>
.. moduleauthor:: Frank Noe <frank DOT noe AT fu-berlin DOT de>

"""

import math
import numpy as np

from .sampler_nrev import SamplerNonRev
from .sampler_rev import SamplerRev
from .sampler_revpi import SamplerRevPi


class TransitionMatrixSampler:

    def __init__(self, count_matrix, reversible=False, mu=None, P0=None, n_steps=1, prior='sparse', seed=-1):

        if not prior == 'sparse':
            raise ValueError("Only Sparse prior is currently implemented")

        self.count_matrix = count_matrix

        # distinguish the sampling cases and initialize accordingly
        if reversible:
            if mu is None:
                if n_steps is None:
                    # use sqrt(n) as a rough guess for the decorrelation time
                    n_steps = math.sqrt(np.shape(count_matrix)[0])
                self.sampler = SamplerRev(count_matrix, P0=P0, seed=seed)
            else:
                if n_steps is None:
                    n_steps = 6  # because we have observed autocorrelation times of about 3.
                self.sampler = SamplerRevPi(count_matrix, mu, P0=P0, seed=seed)
        else:
            if mu is None:
                n_steps = 1  # just force to 1, because this is independent sampling
                self.sampler = SamplerNonRev(count_matrix - 1.0, seed=seed)
            else:
                raise ValueError('Non reversible sampling with fixed stationary vector not implemented')

        # remember number of steps to decorrelate between samples
        self.n_steps = n_steps

    def sample(self, nsamples=1, return_statdist=False, callback=None):
        if nsamples == 1:
            return self.sampler.sample(N=self.n_steps, return_statdist=return_statdist)
        else:
            n = self.count_matrix.shape[0]
            P_samples = np.zeros((nsamples, n, n))
            if return_statdist:
                pi_samples = np.zeros((nsamples, n))
                for i in range(nsamples):
                    P_samples[i, :, :], pi_samples[i, :] = self.sampler.sample(N=self.n_steps, return_statdist=True)
                    if callback is not None:
                        callback()
                return P_samples, pi_samples
            else:
                for i in range(nsamples):
                    P_samples[i, :, :] = self.sampler.sample(N=self.n_steps, return_statdist=False)
                    if callback is not None:
                        callback()
                return P_samples



def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('init', parent_package, top_path)

    config.add_subpackage("discrete")
    config.add_subpackage("gaussian")

    return config



"""This module implements the transition matrix functionality"""

import numpy as np


def log_likelihood(C, T):
    """
        implementation of likelihood of C given T
    """
    C = C.tocsr()
    T = T.tocsr()
    ind = np.nonzero(C)
    relT = np.array(T[ind])[0, :]
    relT = np.log(relT)
    relC = np.array(C[ind])[0, :]

    return relT.dot(relC)



from contextlib import AbstractContextManager
from typing import Optional


def handle_n_jobs(value: Optional[int]) -> int:
    r"""Handles the n_jobs parameter consistently so that a non-negative number is returned.
    In particular, if

      * value is None, use 1 job
      * value is negative, use number cores available * 2
      * value is positive, use value

    Parameters
    ----------
    value : int or None
        The provided n_jobs argument

    Returns
    -------
    n_jobs : int
        A non-negative integer value describing how many threads can be started simultaneously.
    """
    if value is None:
        try:
            from os import sched_getaffinity
            count = len(sched_getaffinity(0))
        except ImportError:
            from os import cpu_count
            count = cpu_count()
        if count is None:
            raise ValueError("Could not determine number of cpus in system, please provide n_jobs manually.")
        value = count
    elif value <= 0:
        raise ValueError(f"n_jobs can only be None (in which case it will be determined from hardware) "
                         f"or a positive number, but was {value}.")
    assert isinstance(value, int) and value > 0
    return value


class joining(AbstractContextManager):
    r""" Context manager for pools that will automatically join upon exit of scope. """

    def __init__(self, thing):
        self.thing = thing

    def __enter__(self):
        return self.thing

    def __exit__(self, *info):
        self.thing.close()
        self.thing.join()



"""This module implements the transition matrix functionality"""

import numpy as np
import scipy.sparse


def transition_matrix_non_reversible(C):
    """implementation of transition_matrix"""
    if not scipy.sparse.issparse(C):
        C = scipy.sparse.csr_matrix(C)
    rowsum = C.tocsr().sum(axis=1)
    # catch div by zero
    if np.min(rowsum) == 0.0:
        raise ValueError("matrix C contains rows with sum zero.")
    rowsum = np.array(1. / rowsum).flatten()
    norm = scipy.sparse.diags(rowsum, 0)
    return norm * C

def correct_transition_matrix(T, reversible=None):
    r"""Normalize transition matrix

    Fixes a the row normalization of a transition matrix.
    To be used with the reversible estimators to fix an almost coverged
    transition matrix.

    Parameters
    ----------
    T : (M, M) ndarray
        matrix to correct
    reversible : boolean
        for future use

    Returns
    -------
    (M, M) ndarray
        corrected transition matrix
    """
    row_sums = T.sum(axis=1).A1
    max_sum = np.max(row_sums)
    if max_sum == 0.0:
         max_sum = 1.0
    return (T + scipy.sparse.diags(-row_sums+max_sum, 0)) / max_sum


def transition_matrix_reversible_pisym(C, return_statdist=False, **kwargs):
    r"""
    Estimates reversible transition matrix as follows:

    ..:math:
        p_{ij} = c_{ij} / c_i where c_i = sum_j c_{ij}
        \pi_j = \sum_j \pi_i p_{ij}
        x_{ij} = \pi_i p_{ij} + \pi_j p_{ji}
        p^{rev}_{ij} = x_{ij} / x_i where x_i = sum_j x_{ij}

    In words: takes the nonreversible transition matrix estimate, uses its
    stationary distribution to compute an equilibrium correlation matrix,
    symmetrizes that correlation matrix and then normalizes to the reversible
    transition matrix estimate.

    Parameters
    ----------
    C: ndarray, shape (n,n)
        count matrix

    Returns
    -------
    T: Estimated transition matrix

    """
    # nonreversible estimate
    T_nonrev = transition_matrix_non_reversible(C)
    from ...analysis import stationary_distribution
    pi = stationary_distribution(T_nonrev)
    # correlation matrix
    X = scipy.sparse.diags(pi).dot(T_nonrev)
    X = X.T + X
    # result
    pi_rev = np.array(X.sum(axis=1)).squeeze()
    T_rev = scipy.sparse.diags(1.0/pi_rev).dot(X)
    if return_statdist:
        return T_rev, pi
    return T_rev



def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration('covariance', parent_package, top_path)
    config.add_subpackage('util')
    config.add_subpackage('util.covar_c')

    return config



r"""Decomposition of a netflux network into its dominant reaction pathways

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>
"""
import warnings
import numpy as np
import scipy.sparse.csgraph as csgraph
from scipy.sparse import coo_matrix


class PathwayError(Exception):
    """Exception for failed attempt to find pathway in a given flux
    network"""
    pass


def find_bottleneck(F, A, B):
    r"""Find dynamic bottleneck of flux network.

    Parameters
    ----------
    F : scipy.sparse matrix
        The flux network
    A : array_like
        The set of starting states
    B : array_like
        The set of end states

    Returns
    -------
    e : tuple of int
    The edge corresponding to the dynamic bottleneck

    """
    if F.nnz == 0:
        raise PathwayError('no more pathways left: Flux matrix does not contain any positive entries')
    F = F.tocoo()
    n = F.shape[0]

    """Get exdges and corresponding flux values"""
    val = F.data
    row = F.row
    col = F.col

    """Sort edges according to flux"""
    ind = np.argsort(val)
    val = val[ind]
    row = row[ind]
    col = col[ind]

    """Check if edge with largest conductivity connects A and B"""
    b = np.array(row[-1], col[-1])
    if has_path(b, A, B):
        return b
    else:
        """Bisection of flux-value array"""
        r = val.size
        l = 0
        N = 0
        while r - l > 1:
            m = np.int64(np.floor(0.5 * (r + l)))
            valtmp = val[m:]
            rowtmp = row[m:]
            coltmp = col[m:]
            C = coo_matrix((valtmp, (rowtmp, coltmp)), shape=(n, n))
            """Check if there is a path connecting A and B by
            iterating over all starting nodes in A"""
            if has_connection(C, A, B):
                l = 1 * m
            else:
                r = 1 * m

        E_AB = coo_matrix((val[l + 1:], (row[l + 1:], col[l + 1:])), shape=(n, n))
        b1 = row[l]
        b2 = col[l]
        return b1, b2, E_AB


def has_connection(graph, A, B):
    r"""Check if the given graph contains a path connecting A and B.

    Parameters
    ----------
    graph : scipy.sparse matrix
        Adjacency matrix of the graph
    A : array_like
        The set of starting states
    B : array_like
        The set of end states

    Returns
    -------
    hc : bool
       True if the graph contains a path connecting A and B, otherwise
       False.

    """
    for istart in A:
        nodes = csgraph.breadth_first_order(graph, istart, directed=True, return_predecessors=False)
        if has_path(nodes, A, B):
            return True
    return False


def has_path(nodes, A, B):
    r"""Test if nodes from a breadth_first_order search lead from A to
    B.

    Parameters
    ----------
    nodes : array_like
        Nodes from breadth_first_oder_seatch
    A : array_like
        The set of educt states
    B : array_like
        The set of product states

    Returns
    -------
    has_path : boolean
        True if there exists a path, else False

    """
    x1 = np.intersect1d(nodes, A).size > 0
    x2 = np.intersect1d(nodes, B).size > 0
    return x1 and x2


def pathway(F, A, B):
    r"""Compute the dominant reaction-pathway.

    Parameters
    ----------
    F : (M, M) scipy.sparse matrix
        The flux network (matrix of netflux values)
    A : array_like
        The set of starting states
    B : array_like
        The set of end states

    Returns
    -------
    w : list
        The dominant reaction-pathway

    """
    if F.nnz == 0:
        raise PathwayError('no more pathways left: Flux matrix does not contain any positive entries')
    b1, b2, F = find_bottleneck(F, A, B)
    if np.any(A == b1):
        wL = [b1, ]
    elif np.any(B == b1):
        raise PathwayError(("Roles of vertices b1 and b2 are switched."
                            "This should never happen for a correct flux network"
                            "obtained from a reversible transition meatrix."))
    else:
        wL = pathway(F, A, [b1, ])
    if np.any(B == b2):
        wR = [b2, ]
    elif np.any(A == b2):
        raise PathwayError(("Roles of vertices b1 and b2 are switched."
                            "This should never happen for a correct flux network"
                            "obtained from a reversible transition meatrix."))
    else:
        wR = pathway(F, [b2, ], B)
    return wL + wR


def capacity(F, path):
    r"""Compute capacity (min. current) of path.

    Paramters
    ---------
    F : (M, M) scipy.sparse matrix
        The flux network (matrix of netflux values)
    path : list
        Reaction path

    Returns
    -------
    c : float
        Capacity (min. current of path)

    """
    F = F.todok()
    L = len(path)
    currents = np.zeros(L - 1)
    for l in range(L - 1):
        i = path[l]
        j = path[l + 1]
        currents[l] = F[i, j]

    return currents.min()


def remove_path(F, path):
    r"""Remove capacity along a path from flux network.

    Parameters
    ----------
    F : (M, M) scipy.sparse matrix
        The flux network (matrix of netflux values)
    path : list
        Reaction path

    Returns
    -------
    F : (M, M) scipy.sparse matrix
        The updated flux network

    """
    c = capacity(F, path)
    F = F.todok()
    L = len(path)
    for l in range(L - 1):
        i = path[l]
        j = path[l + 1]
        F[i, j] -= c
    return F


def pathways(F, A, B, fraction=1.0, maxiter=1000, tol=1e-14):
    r"""Decompose flux network into dominant reaction paths.

    Parameters
    ----------
    F : (M, M) scipy.sparse matrix
        The flux network (matrix of netflux values)
    A : array_like
        The set of starting states
    B : array_like
        The set of end states
    fraction : float, optional
        Fraction of total flux to assemble in pathway decomposition
    maxiter : int, optional
        Maximum number of pathways for decomposition
    tol : float, optional
        Floating point tolerance. The iteration is terminated once the
        relative capacity of all discovered path matches the desired
        fraction within floating point tolerance

    Returns
    -------
    paths : list
        List of dominant reaction pathways
    capacities: list
        List of capacities corresponding to each reactions pathway in paths

    References
    ----------
    .. [1] P. Metzner, C. Schuette and E. Vanden-Eijnden.
        Transition Path Theory for Markov Jump Processes.
        Multiscale Model Simul 7: 1192-1219 (2009)

    """
    F, a, b = add_endstates(F, A, B)
    A = [a, ]
    B = [b, ]

    """Total flux"""
    TF = F.tocsr()[A, :].sum()

    """Total capacity fo all previously found reaction paths"""
    CF = 0.0
    niter = 0

    """List of dominant reaction pathways"""
    paths = []
    """List of corresponding capacities"""
    capacities = []

    while True:
        """Find dominant pathway of flux-network"""
        try:
            path = pathway(F, A, B)
        except PathwayError:
            break
        """Compute capacity of current pathway"""
        c = capacity(F, path)
        """Remove artifical end-states"""
        path = path[1:-1]
        """Append to lists"""
        paths.append(np.array(path))
        capacities.append(c)
        """Update capacity of all previously found paths"""
        CF += c
        """Remove capacity along given path from flux-network"""
        F = remove_path(F, path)
        niter += 1
        """Current flux numerically equals fraction * total flux or is
        greater equal than fraction * total flux"""
        if (abs(CF/TF - fraction) <= tol) or (CF/TF >= fraction):
            break
        if niter > maxiter:
            warnings.warn("Maximum number of iterations reached", RuntimeWarning)
            break
    return paths, capacities


def add_endstates(F, A, B):
    r"""Adds artifical end states replacing source and sink sets.

    Parameters
    ----------
    F : (M, M) scipy.sparse matrix
        The flux network (matrix of netflux values)
    A : array_like
        The set of starting states
    B : array_like
        The set of end states

    Returns
    -------
    F_new : (M+2, M+2) scipy.sparse matrix
        The artifical flux network with extra end states
    a_new : int
        The new single source a_new = M
    b_new : int
        The new single sink b_new = M+1

    """

    """Outgoing currents from A"""
    F = F.tocsr()
    outA = (F[A, :].sum(axis=1)).getA()[:, 0]

    """Incoming currents into B"""
    F = F.tocsc()
    inB = (F[:, B].sum(axis=0)).getA()[0, :]

    F = F.tocoo()
    M = F.shape[0]

    data_old = F.data
    row_old = F.row
    col_old = F.col

    """Add currents from new A=[n,] to all states in A"""
    row1 = np.zeros(outA.shape[0], dtype=int)
    row1[:] = M
    col1 = np.array(A)
    data1 = outA

    """Add currents from old B to new B=[n+1,]"""
    row2 = np.array(B)
    col2 = np.zeros(inB.shape[0], dtype=int)
    col2[:] = M + 1
    data2 = inB

    """Stack data, row and col arrays"""
    data = np.hstack((data_old, data1, data2))
    row = np.hstack((row_old, row1, row2))
    col = np.hstack((col_old, col1, col2))

    """New netflux matrix"""
    F_new = coo_matrix((data, (row, col)), shape=(M + 2, M + 2))

    return F_new, M, M + 1



import numpy as np
from . import _hmm_bindings as _bindings


def observations_in_state(hidden_state_trajectories, observed_state_trajectories, hidden_state):
    dtype = observed_state_trajectories[0].dtype
    collected_observations = np.array([], dtype=dtype)
    for (s_t, o_t) in zip(hidden_state_trajectories, observed_state_trajectories):
        indices = np.where(s_t == hidden_state)[0]
        collected_observations = np.append(collected_observations, o_t[indices])

    # collected_observations = [
    #    o_t[np.where(s_t == state_index)[0]] for s_t, o_t in zip(self.hidden_state_trajectories, observations)
    # ]
    # return np.hstack(collected_observations)
    return collected_observations


def sample_hidden_state_trajectory(transition_matrix, output_model, initial_distribution, obs, temp_alpha=None):
    """Sample a hidden state trajectory from the conditional distribution P(s | T, E, o)

    Parameters
    ----------
    transition_matrix : (n, n) ndarray
        The transition matrix :math:`T` over the hidden states.
    output_model : deeptime.markov.hmm.OutputModel
        Output model with emission probabilities :math:`E`.
    initial_distribution : (n,) ndarray
        Initial distribution over hidden states.
    obs : (T,) ndarray
        Trajectory in observation space.
    temp_alpha : (T, n) ndarray, optional, default=None
        Optional array that is used to store alphas from the forward pass, if provided it is used for storage
        instead of allocating new memory.

    Returns
    -------
    s_t : (T,) ndarray
        Hidden state trajectory, with s_t[t] the hidden state corresponding to observation obs[t]
    """

    # Determine observation trajectory length
    T = obs.shape[0]

    if temp_alpha is None:
        temp_alpha = np.zeros((obs.shape[0], transition_matrix.shape[0]), dtype=transition_matrix.dtype)

    # compute output probability matrix
    pobs = output_model.to_state_probability_trajectory(obs)
    # compute forward variables
    _bindings.util.forward(transition_matrix, pobs, initial_distribution, T=T, alpha_out=temp_alpha)
    # sample path
    S = _bindings.util.sample_path(temp_alpha, transition_matrix, T=T)
    return S



r"""This module implements the connectivity functionality

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>
"""

import numpy as np
import scipy.sparse.csgraph as csgraph

from deeptime.util.matrix import submatrix


def connected_sets(C, directed=True):
    r"""Compute connected components for a directed graph with weights
    represented by the given count matrix.

    Parameters
    ----------
    C : scipy.sparse matrix or numpy ndarray
        square matrix specifying edge weights.
    directed : bool, optional
       Whether to compute connected components for a directed  or
       undirected graph. Default is True.

    Returns
    -------
    cc : list of arrays of integers
        Each entry is an array containing all vertices (states) in
        the corresponding connected component.

    """
    M = C.shape[0]
    """ Compute connected components of C. nc is the number of
    components, indices contain the component labels of the states
    """
    nc, indices = csgraph.connected_components(C, directed=directed, connection='strong')

    states = np.arange(M)  # Discrete states

    """Order indices"""
    ind = np.argsort(indices)
    indices = indices[ind]

    """Order states"""
    states = states[ind]
    """ The state index tuple is now of the following form (states,
    indices)=([s_23, s_17,...,s_3, s_2, ...], [0, 0, ..., 1, 1, ...])
    """

    """Find number of states per component"""
    count = np.bincount(indices)

    """Cumulative sum of count gives start and end indices of
    components"""
    csum = np.zeros(len(count) + 1, dtype=int)
    csum[1:] = np.cumsum(count)

    """Generate list containing components, sort each component by
    increasing state label"""
    cc = []
    for i in range(nc):
        cc.append(np.sort(states[csum[i]:csum[i + 1]]))

    """Sort by size of component - largest component first"""
    cc = sorted(cc, key=lambda x: -len(x))

    return cc


def largest_connected_set(C, directed=True):
    r"""Compute connected components for a directed graph with weights
    represented by the given count matrix.

    Parameters
    ----------
    C : scipy.sparse matrix or numpy ndarray
        Count matrix specifying edge weights.

    Returns
    -------
    lcc : array of integers
        The largest connected component of the directed graph.

    """
    return connected_sets(C, directed=directed)[0]


def largest_connected_submatrix(C, directed=True, lcc=None):
    r"""Compute the count matrix of the largest connected set.

    The input count matrix is used as a weight matrix for the
    construction of a directed graph. The largest connected set of the
    constructed graph is computed. Vertices belonging to the largest
    connected component are used to generate a completely connected
    subgraph. The weight matrix of the subgraph is the desired
    completely connected count matrix.

    Parameters
    ----------
    C : scipy.sparse matrix or numpy ndarray
        Count matrix specifying edge weights
    directed : bool, optional
       Whether to compute connected components for a directed or
       undirected graph. Default is True
    lcc : (M,) ndarray, optional
       The largest connected set

    Returns
    -------
    C_cc : scipy.sparse matrix
        Count matrix of largest completely
        connected set of vertices (states)

    """
    if lcc is None:
        lcc = largest_connected_set(C, directed=directed)
    return submatrix(C, lcc)


def is_connected(C, directed=True):
    r"""Return true, if the input count matrix is completely connected.
    Effectively checking if the number of connected components equals one.

    Parameters
    ----------
    C : scipy.sparse matrix or numpy ndarray
        Count matrix specifying edge weights.
    directed : bool, optional
       Whether to compute connected components for a directed  or
       undirected graph. Default is True.

    Returns
    -------
    connected : boolean, returning true only if C is connected.


    """
    nc = csgraph.connected_components(C, directed=directed, connection='strong', return_labels=False)
    return nc == 1




def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('estimation', parent_package, top_path)
    config.add_subpackage('dense')
    config.add_subpackage('sparse')

    return config
